#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
class options_item:
	def __init__(self,value,description,dtype,name,permissible_values=None,value_description=None,descr_for_vector_setting='',category='General'):
		"""permissible values can be a vector or a string with an inequality, 
		where %s represents the number, for example "1>%s>0"\n
		if permissible_values is a vector, value_description is a corresponding vector with 
		description of each value in permissible_values"""
		#if permissible_values
		self.description=description
		self.value=value
		self.dtype=dtype
		if type(dtype)==str:
			self.dtype_str=dtype
		elif type(dtype)==list or type(dtype)==tuple:
			self.dtype_str=str(dtype).replace('<class ','').replace('[','').replace(']','').replace('>','').replace("'",'')
		else:
			self.dtype_str=dtype.__name__
			
		self.permissible_values=permissible_values
		self.value_description=value_description
		self.descr_for_vector_setting=descr_for_vector_setting
		self.category=category
		self.name=name
		self.selection_var= (descr_for_vector_setting=='') and type(permissible_values)==list
		self.is_inputlist=len(self.descr_for_vector_setting)>0 and type(self.description)==list
	
		
		
	def set(self,value,i=None):
		try:
			if not self.valid(value,i):
				return False
		except Exception as e:
			return False
		if i is None:
			self.value=value
		else:
			if self.value[i]!=value:
				self.value[i]=value
			else:
				return False
		return True
	
	def valid(self,value,i=None):
		if i is None:
			return self.valid_test(value, self.permissible_values)
		else:
			if self.permissible_values is None:
				return (type(value)==self.dtype) or (type(value) in self.dtype)
			return self.valid_test(value, self.permissible_values[i])
			
	def valid_test(self,value,permissible):
		if type(permissible)==list or type(permissible)==tuple:
			try:
				if not type(value)==list or type(value)==tuple:
					value=self.dtype(value)
					return value in permissible
				else:
					valid=True
					for i in range(len(value)):
						value[i]=self.dtype(value[i])
						valid=valid*eval(permissible[i] %(value[i],))
			except:
				return False
			return valid
		elif type(permissible)==str:
			return eval(permissible %(value,))
		else:
			raise RuntimeError('No method to handle this permissible')
		

		
class options():
	def __init__(self):
		pass
		
		
	def make_category_tree(self):
		opt=self.__dict__
		d=dict()
		keys=np.array(list(opt.keys()))
		keys=keys[keys.argsort()]
		for i in opt:
			if opt[i].category in d:
				d[opt[i].category].append(opt[i])
			else:
				d[opt[i].category]=[opt[i]]
			opt[i].code_name=i
		self.categories=d	
		keys=np.array(list(d.keys()))
		self.categories_srtd=keys[keys.argsort()]



def regression_options():
	self=options()
	self.pqdkm						= options_item([1,1,0,1,1], 
																	["Auto Regression order (ARIMA, p)",
																	"Moving Average order (ARIMA, q)",
																	"difference order (ARIMA, d)",
																	"Variance Moving Average order (GARCH, k)",
																	"Variance Auto Regression order (GARCH, m)"],int, 'ARIMA-GARCH orders',
																	["%s>=0","%s>=0","%s>=0","%s>=0","%s>=0"],
																	descr_for_vector_setting="ARIMA-GARCH parameters",category='Regression')
	self.group_fixed_random_eff		= options_item(2,				'Fixed, random or no group effects', int, 'Group fixed random effect',[0,1,2], 
																	['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
	self.time_fixed_random_eff		= options_item(2,				'Fixed, random or no time effects', int, 'Time fixed random effect',[0,1,2], 
																	['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
	self.variance_fixed_random_eff	= options_item(2,				'Fixed, random or no group effects for variance', int, 'Variance fixed random effects',[0,1,2], 
																	['No effects','Fixed effects','Random effects'],category='Fixed-random effects')
	
	self.loadargs					= options_item(1, 				"Determines whether the regression arguments from the previous run should be kept", 
																	int, 'Load arguments', [0,1,2],
																	['No loading',
																	'Load from last run',
																	'Load from last matching model',
																	])
	
	self.loadARIMA_GARCH			= options_item(False, 				"Determines whether the ARIMA_GARCH arguments from the previous run should be kept", 
																	bool, 'Load ARIMA_GARCH', [False,True],
																	['No loading',
																	'Load arguments'
																	])	
	
	self.add_intercept				= options_item(True,			"If True, adds intercept if not all ready in the data",
																	bool,'Add intercept', [True,False],['Add intercept','Do not add intercept'],category='Regression')
	
	self.h_function					= options_item(h_func,			h_descr, str,"GARCH function",category='Regression')
	self.user_constraints			= options_item(None,			constr_str,str, 'User constraints')
	
	self.tobit_limits				= options_item([None,None],		['lower limit','upper limit'], [float,type(None)], 'Tobit-model limits', descr_for_vector_setting=tobit_desc)
	
	self.min_group_df				= options_item(5, 				"The smallest permissible degrees of freedom", int, 'Minimum degrees of freedom', "%s>-1",category='Regression')
	self.robustcov_lags_statistics	= options_item([100,30],		[robust_desc_0,robust_desc_1], int, 'Robust covariance lags (time)', ["%s>1","%s>1"], descr_for_vector_setting=robust_desc_all,category='Output')
	self.silent						= options_item(False, 			silent_desc,  bool,'Silent mode',[True,False],['Silent','Not Silent'])
	self.description				= options_item(None, 			descr_descr, 'entry','Description')	
	
	self.convergence_limit	= options_item([0.0001], ['Convergence limit:'],float,"Convergence limit", ["%s>0"],
									descr_for_vector_setting=conv_desc
									)	
	
	self.make_category_tree()
	
	return self


def application_preferences():
	opt=options()
	
	opt.save_datasets	= options_item(True, "If 1 all loaded datasets are saved on exit and will reappear when the application is restarted", 
									bool,"Save datasets on exit", [False,True],
									['Save on exit',
									'No thanks'])

	opt.make_category_tree()
	
	return opt


h_func="""
def h(e,z):
	e2			=	e**2+1e-5
	h_val		=	np.log(e2)	
	h_e_val		=	2*e/e2
	h_2e_val	=	2/e2-4*e**2/e2**2

	return h_val,h_e_val,h_2e_val,None,None,None
"""

conv_desc="""
Convergence limit. When this exceedes the maximum absolute value of 
new directions (as percentage of parameter value), the procedure is 
determined to have converged"""
h_descr="""
You can supply your own heteroskedasticity function. It must be a function of\n
residuals e and a shift parameter z that is determined by the maximization procedure\n
the function must return the value and its derivatives in the following order:\n
h, dh/de, (d^2)h/de^2, dh/dz, (d^2)h/dz^2,(d^2)h/(dz*de)"""

constr_str="""
You can add constraints in python dictonary syntax.\n
"""

tobit_desc="""
Determines the limits in a tobit regression. 
Element 0 is lower limit and element1 is upper limit. 
If None, the limit is not active"""

robust_desc_all="""
Numer of lags used in calculation of the robust \ncovariance matrix for the time dimension.\n"""
robust_desc_0="""# lags in final statistics calulation       """
robust_desc_1="""# lags iterations (smaller saves time) """

silent_desc="""
True if silent mode, otherwise False. 
For running the procedure in a script, where output should be suppressed"""

descr_descr="""
A description of the project. 
Used in the final output and to load the project later.\n
default is the model_string"""





#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module calculates statistics and saves it to a file



import numpy as np
import stat_functions as stat
from scipy import stats as scstats
import functions as fu
import loglikelihood as logl
from tkinter import font as tkfont
import tkinter as tk
STANDARD_LENGTH=8

class statistics:
	def __init__(self,results,correl_vars=None,descriptives_vars=None,simple_statistics=False,name=None):
		"""This class calculates, stores and prints statistics and statistics"""		

		self.G=results.direction.G
		self.H=results.direction.H
		self.ll=results.ll
		self.panel=results.panel
		self.ll.standardize()
		self.Rsq, self.Rsqadj, self.LL_ratio,self.LL_ratio_OLS=stat.goodness_of_fit(self.panel,self.ll)
		self.LL_restricted=logl.LL(self.panel.args.args_restricted, self.panel).LL
		self.LL_OLS=logl.LL(self.panel.args.args_OLS, self.panel).LL		
		self.name=name
		
		self.no_ac_prob,rhos,RSqAC=stat.breusch_godfrey_test(self.panel,self.ll,10)
		self.norm_prob=stat.JB_normality_test(self.ll.e_st,self.panel)	
		if simple_statistics:	
			self.output=self.arrange_output(results.direction,self.norm_prob,self.no_ac_prob)
			return	
		
		self.output=self.arrange_output(results.direction,self.norm_prob,self.no_ac_prob)
		self.reg_output=self.output.X
		self.multicollinearity_check(self.G)

		self.data_correlations,self.data_statistics=self.correl_and_statistics(correl_vars,descriptives_vars)
		
		self.adf_test=stat.adf_test(self.panel,self.ll,10)
		self.save_stats()
		
	def arrange_output(self,direction,norm_prob,ac_prob):
		panel,H,G,ll=self.panel,self.H,self.G,self.ll
		l=STANDARD_LENGTH
		# python variable name,	lenght ,	is string,		display name,			can take negative values,	justification	next tab space
		pr=[
			['names',			'namelen',	True,			'Variable names',		False,						'right', 		2],
	        ['args',			l,			False,			'Coef',					True,						'right', 		2],
	        ['se_robust',		l,			False,			'robust',				False,						'right', 		2],
	        ['se_st',			l,			False,			'standard',				False,						'right', 		2],
	        ['tstat',			l,			False,			't-stat.',				True,						'right', 		2],
	        ['tsign',			l,			False,			'sign.',				False,						'right', 		1],
	        ['sign_codes',		4,			True,			'',						False,						'left', 		2],
		    ['dx_norm',		    l,			False,			'direction',	    	True,						'right', 		2],
			['multicoll',		1,			True,			'',						False,						'right', 		2],
			['assco',			20,			True,			'associated variable',	False,						'center', 		2],
		    ['set_to',			6,			True,			'restr',			False,							'center', 		2],
		    ['cause',			10,			True,			'cause',				False,						'right', 		2]
		]
		o=output(pr, ll,direction,panel.settings.robustcov_lags_statistics.value[0])
		o.add_heading(top_header=' '*100 + '_'*11+'SE'+'_'*11+" "*73+"_"+"restricted variables"+"_",
		              statistics= [['Normality (Jarque-Bera test for normality)',norm_prob,3,'%'], 
		                           ['Stationarity (Breusch Godfrey_test on AC, significance)',ac_prob,3,'%']])
		o.add_footer("Significance codes: .=0.1, *=0.05, **=0.01, ***=0.001,    |=collinear")
		return o	
	
	def correl_and_statistics(self,correl_vars,descriptives_vars):
		panel=self.panel
		x_names=[]
		X=[]
		correl_X,correl_names=get_variables(panel, correl_vars)
		descr_X,descr_names=get_variables(panel, descriptives_vars)
	

		c=stat.correl(correl_X)
		c=np.concatenate((correl_names,c),0)
		n=descr_X.shape[1]
		vstat=np.concatenate((np.mean(descr_X,0).reshape((n,1)),
		                      np.std(descr_X,0).reshape((n,1)),
		                      np.min(descr_X,0).reshape((n,1)),
		                      np.max(descr_X,0).reshape((n,1))),1)
		vstat=np.concatenate((descr_names.T,vstat),1)
		vstat=np.concatenate(([['','Mean','SD','min','max']],vstat),0)
		correl_names=np.append([['']],correl_names,1).T
		c=np.concatenate((correl_names,c),1)

		return c,vstat
				
	def multicollinearity_check(self,G):
		"Returns a variance decompostition matrix with headings"
		panel=self.panel
		vNames=['Max(var_proportion)','CI:']+panel.args.names_v
		k=len(vNames)-1
		matr=stat.var_decomposition(X=G,concat=True)
		matr=np.round(matr,3)
		maxp=np.max(matr[:,1:],1).reshape((matr.shape[0],1))
		matr=np.concatenate((maxp,matr),1)
		matr=np.concatenate(([vNames],matr))
		self.MultiColl=matr

	def save_stats(self):
		"""Saves the various statistics assigned to self"""
		ll=self.ll
		panel=self.panel
		N,T,k=panel.X.shape
		output=dict()
		name_list=[]
		add_output(output,name_list,'Information',[
		    ['Description:',panel.input.descr],
		    ['LL:',ll.LL],
		    ['Number of IDs:',N],
		    ['Maximum number of dates:',T],
		    ['A) Total number of observations:',panel.NT_before_loss],
		    ['B) Observations lost to GARCH/ARIMA',panel.tot_lost_obs],		
		    ['    Total after loss of observations (A-B):',panel.NT],
		    ['C) Number of Random/Fixed Effects coefficients:',N],
		    ['D) Number of Random/Fixed Effects coefficients in the variance process:',N],
		    ['E) Number of coefficients:',panel.args.n_args],
		    ['DF (A-B-C-D-E):',panel.df],
		    ['RSq:',self.Rsq],
		    ['RSq Adj:',self.Rsqadj],
		    ['LL-ratio:',self.LL_ratio],
		    ['no ac_prob:',self.no_ac_prob],
		    ['norm prob:',self.norm_prob],
		    ['ADF (dicky fuller):',self.adf_test, "1% and 5 % lower limit of confidence intervals, respectively"],
		    ['Dependent:',panel.input.y_name]
		    ])
		
		add_output(output,name_list,'Regression',self.reg_output)
		add_output(output,name_list,'Multicollinearity',self.MultiColl)

		add_output(output,name_list,'Descriptive statistics',self.data_statistics)
		add_output(output,name_list,'Correlation Matrix',self.data_correlations)
		add_output(output,name_list,'Number of dates in each ID',panel.T_arr.reshape((N,1)))
		
		output_table=[['']]
		output_positions=['']
		for i in name_list:
			if i!='Statistics':
				output_table.extend([[''],['']])
			pos=len(output_table)+1
			output_table.extend([[i+':']])
			output_table.extend(output[i])
			output_positions.append('%s~%s~%s~%s' %(i,pos,len(output[i]),len(output[i][0])))
		output_table[0]=output_positions
		if self.name is None:
			fname=panel.input.descr.replace('\n','').replace('\r', '')
		else:
			fname=self.name
		if len(fname)>65:
			fname=fname[:30]+'...'+fname[-30:]
		fu.savevar(output_table,fname+'.csv')
		
		self.output_dict=output
	
def t_stats(args,direction,lags,d):
	d['names']=np.array(direction.panel.args.names_v)
	if direction.H is None:
		return
	d['se_robust'],d['se_st']=sandwich(direction,lags)
	no_nan=np.isnan(d['se_robust'])==False
	valid=no_nan
	valid[no_nan]=(d['se_robust'][no_nan]>0)
	T=len(d['names'])
	d['tstat']=np.array(T*[np.nan])
	d['tsign']=np.array(T*[np.nan])
	d['tstat'][valid]=args[valid]/d['se_robust'][valid]
	d['tsign'][valid]=(1-scstats.t.cdf(np.abs(d['tstat'][valid]),direction.panel.df))#Two sided tests
	d['sign_codes']=get_sign_codes(d['tsign'])

	

def add_variable(name,panel,names,variables):
	if '|~|' in name:
		return
	if name in panel.dataframe.keys():
		d=panel.dataframe[name]
		if type(d)==np.ndarray:
			names.append(name)
			variables.append(d)
			
def get_variables(panel,input_str):
	v=fu.split_input(input_str)
	names=[]
	variables=[]
	if not v is None:
		for i in v:
			add_variable(i, panel, names, variables)
	
	if v is None or len(names)==0:
		for i in panel.dataframe.keys():
			add_variable(i, panel, names, variables)
			
	n=len(names)
	X=np.concatenate(variables,1)
	names=np.array(names).reshape((1,n))
	return X,names
			
def add_output(output_dict,name_list,name,table):
	if type(table)==np.ndarray:
		table=np.concatenate(([[''] for i in range(len(table))],table),1)
	output_dict[name]=table
	name_list.append(name)
	

def get_list_dim(lst):
	"""Returns 0 if not list, 1 if one dim and 2 if two or more dim. If higher than
	2 dim are attemted to print, then each cell will contain an array. Works on lists and ndarray"""
	if  type(lst)==np.ndarray:
		return min((len(lst.shape),2))
	elif type(lst)==list:
		for i in lst:
			if type(i)!=list:
				return 1
		return 2
	else:
		return 0
		
def get_sign_codes(tsign):
	sc=[]
	for i in tsign:
		if np.isnan(i):
			sc.append(i)
		elif i<0.001:
			sc.append('***')
		elif i<0.01:
			sc.append('** ')
		elif i<0.05:
			sc.append('*  ')
		elif i<0.1:
			sc.append(' . ')
		else:
			sc.append('')
	sc=np.array(sc,dtype='<U3')
	return sc

def remove_illegal_signs(name):
	illegals=['#', 	'<', 	'$', 	'+', 
	          '%', 	'>', 	'!', 	'`', 
	          '&', 	'*', 	'‘', 	'|', 
	          '{', 	'?', 	'“', 	'=', 
	          '}', 	'/', 	':', 	
	          '\\', 	'b']
	for i in illegals:
		if i in name:
			name=name.replace(i,'_')
	return name

def constraints_printout(direction,d):
	panel=direction.panel
	constr=direction.constr
	weak_mc_dict=direction.weak_mc_dict
	if not direction.dx_norm is None:
		d['dx_norm']=direction.dx_norm
	T=len(d['names'])
	d['set_to'],d['assco'],d['cause'],d['multicoll']=['']*T,['']*T,['']*T,['']*T
	if constr is None:
		return
	c=constr.fixed
	for i in c:
		d['set_to'][i]=c[i].value_str
		d['assco'][i]=c[i].assco_name
		d['cause'][i]=c[i].cause	
		
	c=constr.intervals
	for i in c:
		if not c[i].intervalbound is None:
			d['set_to'][i]=c[i].intervalbound
			d['assco'][i]='NA'
			d['cause'][i]=c[i].cause		
			
	for i in weak_mc_dict:#adding associates of non-severe multicollinearity
		d['multicoll'][i]='|'
		d['assco'][i]=panel.args.names_v[weak_mc_dict[i][0]]		

class output:
	def __init__(self,pr,ll,direction,lags):
		self.ll=ll
		d={'args':self.ll.args_v}		
		t_stats(d['args'],direction,lags,d)
		constraints_printout(direction,d)	
		self.X,self.tab_spaces=output_matrix(pr, d)		
		p=""
		for i in range(len(self.X)):
			for j in range(len(self.X[0])):
				p+='\t'+self.X[i][j]
			p+='\n'
		self.printstring=p
		self.dict=d
		self.panel=ll.panel
		self.ll=ll
	
	def get_tab_stops(self,font,init_space=1):
		font = tkfont.Font(font=font)
		m_len = font.measure("m")
		counter=init_space*m_len
		tabs=[f"{counter}"]
		n=len(self.X[1])
		for i in range(n):
			counter+=len(self.X[1][i])*m_len
			counter+=self.tab_spaces[i]*m_len
			tabs.extend([f"{counter}"])
		return tabs
		
	def print(self,tab=None):
		if tab is None:
			print(self.printstring)
		else:
			tab.box.replace_all(self.printstring)
	
	def add_heading(self,Iterations=None,top_header=None,statistics=None,incr=None):
		s=("LL: "+str(self.ll.LL)+'  ').ljust(23)
		if not incr is None:
			s+=("Increment: "+ str(incr)).ljust(17)+"  "
		else:
			s+=str(" ").ljust(19)
		if not Iterations is None:
			s+="Iteration: "+ str(Iterations).ljust(7)		
		if not statistics is None:
			for i in statistics:
				if not i[1] is None:
					if i[3]=='%':
						value=str(np.round(i[1]*100,i[2]))+i[3]
						s+=("%s: %s " %(i[0],value)).ljust(16)
					elif i[3]=='decimal': 
						value=np.round(i[1],i[2])
						s+=("%s: %s " %(i[0],value)).ljust(16)
					else:
						s+=str(i[0])+str(i[1])+str(i[3])
					
		s+='\n'
		if not top_header is None:
			s+=top_header+'\n'
		self.printstring=s+self.printstring
			
	def add_footer(self,text):
		self.printstring+='\n'+text
		if True:
			self.printstring+='\n\n'+self.statistics()
		
	def statistics(self):
		panel=self.panel
		ll=self.ll
		ll.standardize()
		N,T,k=panel.X.shape
		Rsq, Rsqadj, LL_ratio,LL_ratio_OLS=stat.goodness_of_fit(panel,ll)	
		no_ac_prob,rhos,RSqAC=stat.breusch_godfrey_test(panel,ll,10)
		norm_prob=stat.JB_normality_test(ll.e_st,panel)

		#Description:{self.panel.input.desc}
		s=f"""
\t                         
\tNumber of IDs                       :{N}
\tNumber of dates (maximum)           :{T}

\tA) Total number of observations     :{panel.NT_before_loss}
\tB) Observations lost to GARCH/ARIMA':{panel.tot_lost_obs}		
\tRandom/Fixed Effects in
\tC)      Mean process                :{panel.number_of_RE_coef}
\tD)      Variance process            :{panel.number_of_RE_coef_in_variance}
\tE) Number of coefficients in regr.  :{panel.args.n_args}
\tDegrees of freedom (A-B-C-D-E)      :{panel.df}

\tR-squared                           :{round(Rsq*100,1)}%
\tAdjusted R-squared                  :{round(Rsqadj*100,1)}%
\tLL-ratio                            :{round(LL_ratio,2)}
\tBreusch-Godfrey-test                :{round(no_ac_prob*100,1)}% (significance, probability of no auto correlation)
\tJarque–Bera test for normality      :{round(norm_prob*100,1)}% (significance, probability of normality)
"""
		ADF_stat,c1,c5=stat.adf_test(panel,ll,10)
		if not ADF_stat=='NA':
			if ADF_stat<c1:
				ADF_res="Unit root rejected at 1%"
			elif ADF_stat<c5:
				ADF_res="Unit root rejected at 5%"
			else:
				ADF_res="Unit root not rejected"		
			adf=f"""
\tAugmented Dicky-Fuller (ADF)        
\t                   Test statistic   :{round(ADF_stat,2)}
\t                   1% critical value:{round(c1,2)}
\t                   5% critical value:{round(c5,2)}
\t                   Result           :{ADF_res}
			"""
		else:
			adf="Unable to calculate ADF"
		if panel.df<1:
			s+="""
WARNING: All your degrees of freedom (df) has been consumed, so statistics cannot be computed.
you can increase df by for example turning off random/fixed effects """
		return s+adf
		
def output_matrix(pr,d):
	"""Formats the output to an output matrix
	pr must have the format directions:\n
	[\n
	is not a string (boolean),
	variable, \n
	column width, \n
	name (string), \n
	correction for negative numbers (boolean),\n
	mid adjusted (True) or left adjusted (False)\n
	]"""	
	T=len(d['names'])
	K=len(pr)
	X=[['   ']+[str(i).rjust(2) for i in range(T)]]
	tab_spaces=[2]
	for a, l,is_string,name,neg,just,sep in pr:
		if a in d:
			if a=='names':
				l=max([len(i) for i in d['names']])
			v=format_var(d[a], l,is_string,name,neg,just,sep)
			X.append([name[:l+sep-1].ljust(l+sep-1)]+list(v))
			tab_spaces.append(sep)
	ret=[[None]*len(X) for i in range(len(X[0]))]
	for i in range(len(X)):#transposing
		for j in range(len(X[0])):
			ret[j][i]=X[i][j]
	return ret,tab_spaces

def format_var(vals,l,is_string,name,neg,just,sep):
	v=np.array(vals).astype('<U128')	
	if is_string:
		for i in range(len(v)):
			if just=='center':
				v[i]=v[i].center(l)[:l]
			elif just=='right':
				v[i]=v[i].rjust(l)[:l]
			else:
				v[i]=v[i].ljust(l)[:l]	
	else:
		max_l=0
		a=[None]*len(v)
		for i in range(len(v)):
			if 'e' in v[i]:
				v[i]=np.format_float_scientific(vals[i],l-7)
			a[i]=v[i].split('.')
			max_l=max((max_l,len(a[i][0])))			
		if max_l>l-3:
			for i in range(len(v)):
				v[i]=np.format_float_scientific(vals[i],l-3)
			return v
		for i in range(len(v)):
			if len(a[i])==2:
				v[i]=(a[i][0].rjust(max_l) + '.' + a[i][1])[:l].ljust(l)
			else:
				v[i]=v[i][:l].center(l)

	
	return v
	

def sandwich(direction,lags):
	panel=direction.panel
	H,G,delmap,idx=reduce_size(direction)
	lags=lags+panel.lost_obs
	hessin=np.linalg.inv(-H)
	se_robust,se=stat.robust_se(panel,lags,hessin,G)
	se_robust,se=expand_x(se_robust, idx),expand_x(se, idx)
	return se_robust,se

def reduce_size(direction):
	H=direction.H
	G=direction.G
	if (G is None) or (H is None):
		return
	weak_mc_dict=direction.weak_mc_dict.keys()
	constr=list(direction.constr.fixed.keys())
	for i in weak_mc_dict:
		if not i in constr:
			constr.append(i)
	m=len(H)
	idx=np.ones(m,dtype=bool)
	delmap=np.arange(m)
	if len(constr)>0:#removing fixed constraints from the matrix
		idx[constr]=False
		H=H[idx][:,idx]
		G=G[:,:,idx]
		delmap-=np.cumsum(idx==False)
		delmap[idx==False]=m#if for some odd reason, the deleted variables are referenced later, an out-of-bounds error is thrown	
	return H,G,delmap,idx

def expand_x(x,idx):
	m=len(idx)
	x_full=np.zeros(m)
	x_full[:]=np.nan
	x_full[idx]=x
	return x_full
	
	
	


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import stat_functions as stat
import functions as fu

MAX_COLLINEARITY=1e+18
SMALL_COLLINEARITY=30

def add_static_constraints(constr,panel,ll,its):
	
	add_custom_constraints(constr,panel.settings.user_constraints.value,ll)
	general_constraints=[('rho',-2,2),('lambda',-2,2),('gamma',-2,2),('psi',-2,2)]
	add_custom_constraints(constr,general_constraints,ll)
	p,q,d,k,m=panel.pqdkm
	
	if panel.m_zero:
		constr.add(panel.args.positions['psi'][0],None,'GARCH input constr',value=0.05)
	fix_gamma=(k*m>0)
	if not ll is None and fix_gamma:
		fix_gamma=ll.args_d['psi'][0]==0
	if fix_gamma and (not panel.settings.loadargs.value):
		for i in panel.args.positions['gamma']:
			constr.add(i,None,'GARCH term cannot be positive if ARCH term is zero',value=0)
		

	
	
def add_dynamic_constraints(ll, direction,newton_failed):

	weak_mc_dict=remove_all_multicoll(direction,ll,newton_failed)
	remove_singularities(direction.constr, direction.H)
	return weak_mc_dict


	
def remove_singularities(constr,hessian):
	habs=np.abs(hessian)
	sing_problems=[]
	try:
		a=np.linalg.det(hessian)
		if a>0:
			return sing_problems
	except:
		for i in np.nonzero(np.diag(habs)>1e+100)[0]:
			constr.add(i,None,'singularity (extreeme value in diagonal)')
	for i in np.nonzero(np.diag(hessian)==0)[0]:
		constr.add(i,None,'singularity (zero in diagonal)')
		

def add_custom_constraints(constr,constraints,ll):
	"""Adds custom range constraints\n\n
		constraints shall be on the format [(name, minimum, maximum), ...]"""	
	if constraints is None:
		return
	for i in constraints:
		add_custom_constraint(constr,i,ll)


def add_custom_constraint(constr,constraint,ll):
	"""Adds a custom range constraint\n\n
		constraint shall be on the format (name, minimum, maximum)"""
	name, minimum, maximum=constraint
	m=[minimum,maximum]
	for i in range(2):
		if type(m[i])==str:
			if not ll is None:
				m[i]=eval(m[i],globals(),ll.__dict__)
			else:
				return
	[minimum,maximum]=m
	constr.add_named(name,None,'user constraint', [minimum,maximum])



class constraint:
	def __init__(self,index,assco,cause,value, interval,panel,category):
		names=panel.args.names_v
		self.name=names[index]
		self.intervalbound=None
		self.max=None
		self.min=None
		self.value=None
		self.value_str=None
		if interval is None:
			self.value=value
			self.value_str=str(round(self.value,8))
		else:
			if interval[0]>interval[1]:
				raise RuntimeError('Lower constraint cannot exceed upper')
			self.min=interval[0]
			self.max=interval[1]
			self.cause='user/general constraint'
		self.assco_ix=assco
		if assco is None:
			self.assco_name=None
		else:
			self.assco_name=names[assco]
		self.cause=cause
		self.category=category	
		
class constraints:

	"""Stores the constraints of the LL maximization"""
	def __init__(self,panel,args):
		self.names=panel.args.names_v
		self.panel=panel
		self.categories={}
		self.fixed={}
		self.intervals={}
		self.constraints={}
		self.associates={}
		self.collinears={}
		self.args=args
		self.CI=None


	def add(self,index,assco,cause,interval=None,replace=True,value=None):
		"""Adds a constraint. 'index' is the position
		for which the constraints shall apply.  \n\n

		Equality constraints are chosen by specifying 'minimum_or_value' \n\n
		Inequality constraints are chosen specifiying 'maximum' and 'minimum'\n\n
		'replace' determines whether an existing constraint shall be replaced or not 
		(only one equality and inequality allowed per position)"""


		if ((not replace) and (index in self.constraints)):
			return False
		if interval is None:#this is a fixed constaint
			if value is None:
				value=self.args[index]
			if index in self.intervals:
				c=self.constraints[index]
				if not (c.min<value<c.max):
					return False
				else:
					self.intervals.pop(index)
		elif index in self.fixed: #this is an interval constraint
			self.fixed.pop(index)
			
		eq,category,j=self.panel.args.positions_map[index]
		if not category in self.categories:
			self.categories[category]=[index]
		elif not index in self.categories[category]:
			self.categories[category].append(index)

		c=constraint(index,assco,cause,value, interval ,self.panel,category)
		self.constraints[index]=c
		if value is None:
			self.intervals[index]=c
		else:
			self.fixed[index]=c
		if not assco is None:
			if not assco in self.associates:
				self.associates[assco]=[index]
			elif not index in self.associates[assco]:
				self.associates[assco].append(index)
		if cause=='collinear':
			self.collinears[index]=assco
			
			
				

		return True
	
	def reset_collinears(self,new_args):
		for i in self.collinears:
			self.constraints[i].value=new_args[i]
		
	def delete(self,index):
		if not index in self.constraints:
			return False
		self.constraints.pop(index)
		if index in self.intervals:
			self.intervals.pop(index)
		if index in self.fixed:
			self.fixed.pop(index)		
		eq,category,j=self.panel.args.positions_map[index]
		c=self.categories[category]
		if len(c)==1:
			if c[0]!=index:
				raise RuntimeError('this does not make sense')
			self.categories.pop(category)
		else:
			i=np.nonzero(np.array(c)==index)[0][0]
			c.pop(i)
		a=self.associates
		for i in a:
			if index in a[i]:
				if len(a[i])==1:
					a.pop(i)
					break
				else:
					j=np.nonzero(np.array(a[i])==index)[0][0]
					a[i].pop(j)
		if index in self.collinears:
			self.collinears.pop(index)
		return True
		

	def add_named(self,name,assco,cause,interval):
		positions=self.panel.args.positions[name]
		for i in positions:
			if interval[1] is None:
				self.add(i,assco,cause, value=interval[0])
			else:
				self.add(i,assco,cause, interval)
		return
	
	def set_fixed(self,x):
		"""Sets all elements of x that has fixed constraints to the constraint value"""
		for i in self.fixed:
			x[i]=self.fixed[i].value
		
	def within(self,x,fix=False):
		"""Chekcs if x is within interval constraints. If fix=True, then elements of
		x outside constraints are set to the nearest constraint. if fix=False, the function 
		returns False if x is within constraints and True otherwise"""
		for i in self.intervals:
			c=self.intervals[i]
			if (c.min<=x[i]<=c.max):
				c.intervalbound=None
			else:
				if fix:
					x[i]=max((min((x[i],c.max)),c.min))
					c.intervalbound=str(round(x[i],8))
				else:
					return False
				
		return True
	
	


def append_to_ID(ID,intlist):
	inID=False
	for i in intlist:
		if i in ID:
			inID=True
			break
	if inID:
		for j in intlist:
			if not j in ID:
				ID.append(j)
		return True
	else:
		return False

def correl_IDs(p):
	IDs=[]
	appended=False
	x=np.array(p[:,1:3],dtype=int)
	for i,j in x:
		for k in range(len(IDs)):
			appended=append_to_ID(IDs[k],[i,j])
			if appended:
				break
		if not appended:
			IDs.append([i,j])
	g=len(IDs)
	keep=g*[True]
	for k in range(g):
		if keep[k]:
			for h in range(k+1,len(IDs)):
				appended=False
				for m in range(len(IDs[h])):
					if IDs[h][m] in IDs[k]:
						appended=append_to_ID(IDs[k],  IDs[h])
						keep[h]=False
						break
	g=[]
	for i in range(len(IDs)):
		if keep[i]:
			g.append(IDs[i])
	return g

def normalize(H,include):
	C=-H[include][:,include]
	d=np.maximum(np.diag(C).reshape((len(C),1)),1e-30)**0.5
	C=C/(d*d.T)
	includemap=np.arange(len(include))[include]
	return C,includemap
	
def decomposition(H,include):
	C,includemap=normalize(H, include)
	c_index,var_prop=stat.var_decomposition(XXNorm=C)
	try:
		c_index,var_prop=stat.var_decomposition(XXNorm=C)
	except:
		return None,None,None
	c_index=c_index.flatten()
	return c_index, var_prop,includemap
	
	
def multicoll_problems(direction):
	constr=direction.constr
	H=direction.H
	k,k=H.shape
	include=np.array(k*[True])
	include[list(constr.fixed)]=False
	c_index, var_prop, includemap = decomposition(H, include)
	if c_index is None:
		return False
	constr.CI=c_index[-1]
	mc_problems=[]#list of [index,associate,condition index]
	for cix in range(1,len(c_index)):
		if c_index[-cix]>SMALL_COLLINEARITY:
			if np.sum(var_prop[-cix]>0.5)>1:
				var_prop_ix=np.argsort(var_prop[-cix])[::-1]
				var_prop_val=var_prop[-cix][var_prop_ix]
				j=var_prop_ix[0]
				j=includemap[j]
				done=var_prop_check(direction.panel,var_prop_ix, var_prop_val, includemap,j,mc_problems,c_index[-cix])
				if done:
					break
	return mc_problems

def var_prop_check(panel,var_prop_ix,var_prop_val,includemap,assc,mc_problems,cond_index):
	for i in range(1,len(var_prop_ix)):
		if var_prop_val[i]<0.5:
			return True
		index=var_prop_ix[i]
		index=includemap[index]
		mc_problems.append([index,assc,cond_index])
		return False
		
def add_mc_constraint(direction,mc_problems,weak_mc_dict,newton_failed):
	"""Adds constraints for severe MC problems"""
	constr=direction.constr
	if newton_failed:
		limit=SMALL_COLLINEARITY
	else:
		limit=MAX_COLLINEARITY
	if len(mc_problems)==0:
		return False
	a=np.array(mc_problems)
	a=np.argsort(a[:,2])[::-1]
	for i in a:
		index,assc,cond_index=mc_problems[i]
		if (not index in weak_mc_dict) and (cond_index>SMALL_COLLINEARITY) and (cond_index<=MAX_COLLINEARITY):
			weak_mc_dict[index]=[assc,cond_index]#contains also collinear variables that are only slightly collinear, which shall be restricted when calcuating CV-matrix.	
		if not ((index in constr.associates) or (index in constr.collinears)) and cond_index>limit:
			constr.add(index,assc,'collinear')
			return True
	return False
		
		
def remove_all_multicoll(direction,ll,newton_failed):
	H=direction.H
	k,k=H.shape
	weak_mc_dict=dict()
	for i in range(k):
		mc_problems=multicoll_problems(direction)
		remvd=add_mc_constraint(direction,mc_problems,weak_mc_dict,newton_failed)
		if not remvd:
			break
	select_arma(direction.constr, ll)
	return weak_mc_dict



def select_arma(constr,ll):
	for i in constr.collinears:
		if constr.constraints[i].category in ['rho','lambda', 'gamma','psi']:
			assc=constr.collinears[i]
			if not assc in constr.fixed:
				if abs(ll.args_v[assc])<abs(ll.args_v[i]):
					constr.delete(i)
					constr.add(assc,i,'collinear')
	
	
			
		





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
sys.path.append(__file__.replace("__init__.py",''))
#import system_main as main
import main
import sim_module
import functions as fu
from gui import gui
import options as opt_module
import inspect
import numpy as np

#Todo: check that if works for no id and date variable
#add argument for null model (default: Y~Intercept)


def start():
	"""Starts the GUI"""
	window=gui.window()
	window.mainloop() 

def execute(model_string,dataframe, ID=None,T=None,HF=None):
	"""optimizes LL using the optimization procedure in the maximize module"""
	
	window=main.identify_global(inspect.stack()[1][0].f_globals,'window')
	exe_tab=main.identify_global(inspect.stack()[1][0].f_globals,'exe_tab')
	r=main.execute(model_string,dataframe,ID, T,HF,options,window,exe_tab)
	return r

def statistics(results,correl_vars=None,descriptives_vars=None,name=None):
	return main.output.statistics(results,correl_vars,descriptives_vars,name)

def load_json(fname):

	if False:#detects previously loaded dataset in the environment
		dataframe=main.indentify_dataset(globals(),fname)
		if (not dataframe==False) and (not dataframe is None):
			return dataframe	
	try:
		dataframe=main.loaddata.load_json(fname)
	except FileNotFoundError:
		raise RuntimeError("File %s not found" %(fname))
	main.model_parser.modify_dataframe(dataframe)
	return dataframe


def load(fname,sep=None,dateformat='%Y-%m-%d',load_tmp_data=False):

	"""Loads data from file <fname>, asuming column separator <sep>.\n
	Returns a dataframe (a dictionary of numpy column matrices).\n
	If sep is not supplied, the method will attemt to find it."""
	if False:#detects previously loaded dataset in the environment
		dataframe=main.indentify_dataset(globals(),fname)
		if (not dataframe==False) and (not dataframe is None):
			return dataframe	
	try:
		dataframe=main.loaddata.load(fname,sep,dateformat,load_tmp_data)
	except FileNotFoundError:
		raise RuntimeError("File %s not found" %(fname))
	main.model_parser.modify_dataframe(dataframe)
	return dataframe

def load_SQL(conn,sql_string,dateformat='%Y-%m-%d',load_tmp_data=True):

	"""Loads data from an SQL server, using sql_string as query"""
	if False:#detects previously loaded dataset in the environment
		dataframe=main.indentify_dataset(globals(),sql_string)
		if (not dataframe==False) and (not dataframe is None):
			return dataframe
	dataframe=main.loaddata.load_SQL(conn,sql_string,dateformat,load_tmp_data)
	#except RuntimeError as e:
	#	raise RuntimeError(e)
	main.model_parser.modify_dataframe(dataframe)
	return dataframe

def filter_data(data_filters,dataset):
	d,n=main.model_parser.filter_data(data_filters,dataset)
	return d

def edit_data(edit_script,dataset):
	exec(edit_script,dataset,dataset)
	for i in list(dataset.keys()):
		if not type(dataset[i])==np.ndarray:
			dataset.pop(i)
	window=main.identify_global(inspect.stack()[1][0].f_globals,'window')
	if hasattr(dataset,'datasets') and (not window is None):
		dataset.datasets.make_tree(dataset.name,window.right_tabs.data_tree)	
		
options=opt_module.regression_options()
preferences=opt_module.application_preferences()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

#contains the log likelihood object
import sys
#sys.path.append(__file__.replace("paneltime\\loglikelihood.py",'build\\lib.win-amd64-3.5'))
#sys.path.append(__file__.replace("paneltime\\loglikelihood.py",'build\\lib.linux-x86_64-3.5'))
try:#only using c function if installed
	import cfunctions as c
except ImportError as e:
	c=None
import numpy as np
import functions as fu
import calculus_functions as cf
import stat_functions as stat
import random_effects as re
from scipy import sparse as sp
import scipy
import debug
import traceback

class LL:
	"""Calculates the log likelihood given arguments arg (either in dictonary or array form), and creates an object
	that store dynamic variables that depend on the \n
	If args is a dictionary, the ARMA-GARCH orders are 
	determined from the dictionary. If args is a vector, the ARMA-GARCH order needs to be consistent
	with the  panel object
	"""
	def __init__(self,args,panel,constraints=None,print_err=False):
		self.err_msg=''
		self.errmsg_h=''
		self.panel=panel
		gfre=panel.settings.group_fixed_random_eff.value
		tfre=panel.settings.time_fixed_random_eff.value
		vfre=panel.settings.variance_fixed_random_eff.value
		self.re_obj_i=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre)
		self.re_obj_t=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre)
		self.re_obj_i_v=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre*vfre)
		self.re_obj_t_v=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre*vfre)
		
		self.LL_const=-0.5*np.log(2*np.pi)
	
		self.args_v=panel.args.conv_to_vector(args)
		if not constraints is None:
			constraints.within(self.args_v,True)	
			constraints.set_fixed(self.args_v)
		self.args_d=panel.args.conv_to_dict(self.args_v)
		self.h_err=""
		self.LL=None
		#self.LL=self.LL_calc()
		try:
			self.LL=self.LL_calc()
			if np.isnan(self.LL):
				self.LL=None						
		except Exception as e:
			if print_err:
				traceback.print_exc()
				print(self.errmsg_h)
		
		


	def LL_calc(self):
		panel=self.panel
		X=panel.X
		matrices=set_garch_arch(panel,self.args_d)
		if matrices is None:
			return None		
		
		AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
		(N,T,k)=X.shape

		u = panel.Y-cf.dot(X,self.args_d['beta'])
		e = cf.dot(AMA_1AR,u)
		e_RE = (e+self.re_obj_i.RE(e)+self.re_obj_t.RE(e))*panel.included
		e_REsq = e_RE**2		

		lnv_ARMA = self.garch(GAR_1MA, e)
		W_omega = cf.dot(panel.W_a, self.args_d['omega'])
		lnv = W_omega+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'
		#self.lnv0=lnv*1#debug
		grp = self.variance_RE(e_REsq)
		lnv+=grp
		self.dlnv_pos=(lnv<100)*(lnv>-100)
		lnv = np.maximum(np.minimum(lnv,100),-100)
		v = np.exp(lnv)*panel.a
		v_inv = np.exp(-lnv)*panel.a	
		
		LL = self.LL_const-0.5*(lnv+(e_REsq)*v_inv)
		#self.LL_array=LL#debug
		self.add_variables(matrices, u, e, lnv_ARMA, lnv, v, W_omega, grp,e_RE,e_REsq,v_inv)
		self.tobit(panel,LL)

		LL=np.sum(LL*panel.included)
		if abs(LL)>1e+100: 
			return None				
		return LL
		
	def add_variables(self,matrices,u,e,lnv_ARMA,lnv,v,W_omega,grp,e_RE,e_REsq,v_inv):
		self.e_st=e_RE*v_inv**0.5	
		self.e_st_centered=(self.e_st-self.panel.mean(self.e_st))*self.panel.included
		self.AMA_1,self.AMA_1AR,self.GAR_1,self.GAR_1MA=matrices
		self.u,self.e, self.lnv_ARMA        = u,         e,       lnv_ARMA
		self.lnv,self.v                     = lnv,       v
		self.W_omega=W_omega
		self.grp=grp
		self.e_RE=e_RE
		self.e_REsq=e_REsq
		self.v_inv=v_inv

	
	def tobit(self,panel,LL):
		if sum(panel.input.tobit_active)==0:
			return
		g=[1,-1]
		self.F=[None,None]	
		for i in [0,1]:
			if panel.input.tobit_active[i]:
				I=panel.tobit_I[i]
				self.F[i]= scipy.stats.norm.cdf(g[i]*self.e_st[I])
				LL[I]=np.log(self.F[i])

	def garch(self,GAR_1MA,e):
		if self.panel.pqdkm[4]>0:
			if self.panel.z_active:
				h_res=self.h(e, self.args_d['z'][0])
			else:
				h_res=self.h(e, None)
			(self.h_val,     self.h_e_val,
			 self.h_2e_val,  self.h_z_val,
			 self.h_2z_val,  self.h_ez_val)=[cf.prod((i,self.panel.included)) for i in h_res]
			return cf.dot(GAR_1MA,self.h_val)
		else:
			(self.h_val,    self.h_e_val,
			 self.h_2e_val, self.h_z_val,
			 self.h_2z_val, self.h_ez_val,
			 self.avg_h)=(0,0,0,0,0,0,0)
			return 0			
	
	def variance_RE(self,e_REsq):

		self.vRE,self.lnvRE,self.dlnvRE=0,0,0
		self.ddlnvRE,self.dlnvRE_mu,self.ddlnvRE_mu_vRE=0,None,None
		self.varRE_input, self.ddvarRE_input, self.dvarRE_input = None, None, None
		if self.panel.settings.group_fixed_random_eff.value==0:
			return 0
		panel=self.panel
		if panel.N==0:
			return None

		meane2=panel.mean(e_REsq)
		self.varRE_input=(e_REsq-meane2)*panel.included

		mine2=1e-10
		mu=0.00001
		self.vRE_i=self.re_obj_i_v.RE(self.varRE_input)
		self.vRE_t=self.re_obj_t_v.RE(self.varRE_input)
		self.meane2=meane2
		vRE=meane2*panel.included-self.re_obj_i_v.RE(self.varRE_input)-self.re_obj_t_v.RE(self.varRE_input)
		self.vRE=vRE
		small=vRE<=mine2
		big=small==False
		vREbig=vRE[big]
		vREsmall=vRE[small]

		lnvREbig=np.log(vREbig+mu)
		lnvREsmall=(np.log(mine2+mu)-1+(1/(mine2+mu))*vREsmall)
		lnvRE,dlnvRE,ddlnvRE=np.zeros(vRE.shape),np.zeros(vRE.shape),np.zeros(vRE.shape)
		
		lnvRE[big]=lnvREbig
		lnvRE[small]=lnvREsmall
		self.lnvRE=lnvRE*panel.included

		dlnvRE[big]=1/(vREbig+mu)
		dlnvRE[small]=1/(mine2+mu)
		self.dlnvRE=dlnvRE*panel.included
		
		ddlnvRE[big]=-1/(vREbig+mu)**2
		self.ddlnvRE=ddlnvRE*panel.included
	
		return self.lnvRE
		


	def standardize(self):
		"""Adds X and Y and error terms after ARIMA-E-GARCH transformation and random effects to self"""
		if hasattr(self,'Y_st'):
			return		
		sd_inv=self.v_inv**0.5
		panel=self.panel
		m=panel.lost_obs
		N,T,k=panel.X.shape
		if 'Intercept' in panel.args.names_d['beta']:
			m=self.args_d['beta'][0,0]
		else:
			m=panel.mean(panel.Y)	
		Y=cf.dot(self.AMA_1AR,panel.Y)
		Y=(Y+self.re_obj_i.RE(Y,False)+self.re_obj_t.RE(Y,False))*sd_inv
		X=cf.dot(self.AMA_1AR,panel.X)
		X=(X+self.re_obj_i.RE(X,False)+self.re_obj_t.RE(X,False))*sd_inv
		self.Y_st=Y
		self.X_st=X
		self.Y_pred_st=cf.dot(X,self.args_d['beta'])
		incl=panel.included.reshape(N,T)
		self.e_st_long=self.e_st[incl,:]
		self.Y_st_long=self.Y_st[incl,:]
		self.X_st_long=self.X_st[incl,:]

	def copy_args_d(self):
		return fu.copy_array_dict(self.args_d)

	
	def h(self,e,z):
		try:
			d=dict()
			exec(self.panel.h_def,globals(),d)
			return d['h'](e,z)
		except Exception as err:
			if self.h_err!=str(err):
				self.errmsg_h="Warning,error in the ARCH error function h(e,z): %s" %(err)
			h_err=str(e)


def set_garch_arch(panel,args):
	if c is None:
		m=set_garch_arch_scipy(panel,args)
	else:
		m=set_garch_arch_c(panel,args)
	return m
		
		
def set_garch_arch_c(panel,args):
	"""Solves X*a=b for a where X is a banded matrix with 1 or zero, and args along
	the diagonal band"""
	n=panel.max_T
	rho=np.insert(-args['rho'],0,1)
	psi=np.insert(args['psi'],0,0)

	r=np.arange(n)
	AMA_1,AMA_1AR,GAR_1,GAR_1MA=(
	    np.diag(np.ones(n)),
		np.zeros((n,n)),
		np.diag(np.ones(n)),
		np.zeros((n,n))
	)
	c.bandinverse(args['lambda'],rho,-args['gamma'],psi,n,AMA_1,AMA_1AR,GAR_1,GAR_1MA)
	return  AMA_1,AMA_1AR,GAR_1,GAR_1MA


def set_garch_arch_scipy(panel,args):

	p,q,d,k,m=panel.pqdkm
	nW,n=panel.nW,panel.max_T

	AAR=-lag_matr(-panel.I,args['rho'])
	AMA_1AR,AMA_1=solve_mult(args['lambda'], AAR, panel.I)
	if AMA_1AR is None:
		return
	GMA=lag_matr(panel.zero,args['psi'])	
	GAR_1MA,GAR_1=solve_mult(-args['gamma'], GMA, panel.I)
	if GAR_1MA is None:
		return
	return AMA_1,AMA_1AR,GAR_1,GAR_1MA
	
def solve_mult(args,b,I):
	"""Solves X*a=b for a where X is a banded matrix with 1  and args along
	the diagonal band"""
	n=len(b)
	q=len(args)
	X=np.zeros((q+1,n))
	X[0,:]=1
	X2=np.zeros((n,n))
	w=np.zeros(n)
	r=np.arange(n)	
	for i in range(q):
		X[i+1,:n-i-1]=args[i]
	try:
		X_1=scipy.linalg.solve_banded((q,0), X, I)
		if np.any(np.isnan(X_1)):
			return None,None			
		X_1b=cf.dot(X_1, b)
	except:
		return None,None

	return X_1b,X_1


def add_to_matrices(X_1,X_1b,a,ab,r):
	for i in range(0,len(a)):	
		if i>0:
			d=(r[i:],r[:-i])
			X_1[d]=a[i]
		else:
			d=(r,r)
		X_1b[d]=ab[i]	
	return X_1,X_1b

def lag_matr(L,args):
	k=len(args)
	if k==0:
		return L
	L=1*L
	r=np.arange(len(L))
	for i in range(k):
		d=(r[i+1:],r[:-i-1])
		L[d]=args[i]

	return L


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import functions as fu
import date_time
from datetime import datetime
from datetime import date
import tempstore
NON_NUMERIC_TAG='|~|'
forbidden_names=['tobit_low','tobit_high','Intercept']

def load(fname,sep,dateformat,load_tmp_data):
	fname=fu.obtain_fname(fname)
	if load_tmp_data:
		data=tempstore.loaddata(fname)
		if not data is None:
			load_data_printout(data)
			return data
	heading,s=get_head_and_sep(fname,sep)
	print ("opening file ...")
	data=np.loadtxt(fname,delimiter=s,skiprows=1,dtype=bytes)
	data=data.astype(str)
	print ("... done")
	data=convert_to_numeric_dict(data,heading,dateformat)
	tempstore.savedata(fname,data)
	load_data_printout(data)
	return data

def load_json(fname):
	file=open(fname,encoding='utf-8')
	data=json.load(file)
	df=dict()	
	for r in data:
		for k in r:
			if type(r[k])==dict():
				for m in r[k]:
					key=k+'_'+m
					append(df, key, r[k][m])
			else:
				append(df, k, r[k])
	return df
						
def append(d,key,i):
	if key in d:
		d[key].append(i)
	else:
		d[key]=[i]

def load_SQL(conn,sql_string,dateformat,load_tmp_data):
	if load_tmp_data:
		data=tempstore.loaddata(sql_string)
		if not data is None:
			print('local data loaded')
			load_data_printout(data)
			return data
	crsr=conn.cursor()
	print ("fetching SQL data ...")
	crsr.execute(sql_string)
	data=np.array(crsr.fetchall())
	print ("... done")
	heading=[]
	dtypes=[]
	for i in crsr.description:
		heading.append(i[0])
		if i[1] in SQL_type_dict:
			dtypes.append(i[1])
		else:
			dtypes.append(None)
	data=convert_to_numeric_dict(data,heading,dateformat,dtypes)
	remove_nan(data)
	tempstore.savedata(sql_string,data)
	load_data_printout(data)
	return data

def load_data_printout(data):
	lst=[]
	for i in data:
		if not NON_NUMERIC_TAG in i:
			lst.append(i)
	print ("The following variables were loaded:"+', '.join(lst))
	
def remove_nan(data):
	#Todo: add functionality to delete variables that cause too many deletions
	lst=list(data.keys())
	
	for k in lst:
		try:
			notnan=(np.isnan(data[k])==0)
			break
		except:
			pass
	for i in data:
		if not NON_NUMERIC_TAG in i:
			try:
				isnan=np.isnan(data[i])
				if not sum(isnan)>int(0.5*len(data[i])):
					notnan=(notnan*(isnan==0))
			except:
				pass
	for i in data:
		data[i]=data[i][notnan]
	print("%s observations removed because they were nan" %(len(notnan)-np.sum(notnan)))
		


	

	
	
	
def get_name(x,x_names,default):
	x=get_names(x,x_names)
	if x==[]:
		return default
	else:
		return x[0]
	
def get_names(x,x_names):
	if x_names is None:
		if x is None:
			return []
		else:
			x_names=x
	if type(x_names)==list or type(x_names)==tuple or type(x_names)==np.ndarray:
		if type(x_names[0])==str:
			return list(x_names)
		else:
			raise RuntimeError("Variable names need to be string, list or tuples. Type %s cannot be used" %(type(x_names[0])))
	elif type(x_names)==str: 
		if ',' in x_names and '\n' in x_names:
			raise RuntimeError("X-variables needs to be either comma or linfeed separated. You cannot use both")
		for s in [',','\n',' ','\t']:
			if  s in x_names:
				return fu.clean(x_names,s)
			if s=='\t':#happens when used delimiter was not found
				return [fu.clean(x_names)]#allows for the possibilty of a single variable
	else:
		raise RuntimeError("Variable names need to be string, list or tuples. Type %s cannot be used" %(type(x_names)))


def is_number(s):
	try:
		float(s)
		return True
	except ValueError:
		return False
	
def convert_to_numeric_dict(data,names,dateformat,dtypes=None):
	N,k=data.shape
	df=dict()
	if dtypes is None:
		dtypes=k*[None]
	for i in range(k):
		name=names[i]
		if name in forbidden_names:
			print(f"You can't call a variable {name}, since it is in use by paneltime. The variable is renamed {name}_" )
			name=name+'_'
		make_numeric(data[:,i:i+1],name,df,dateformat,dtypes[i])	
	return df
	
def make_numeric(variable,name,df,dateformat,dtype):
	if not dtype is None and dtype in SQL_type_dict:
		try:
			df[name]=np.array(variable,dtype=SQL_type_dict[dtype])
			return
		except:
			pass
	try:
		try_float_int(variable, df, name)
	except:
		try:
			check_dateness(variable,df,name,dateformat)
		except ValueError:
			convert_cat_to_int(variable,df,name)
		df[name+NON_NUMERIC_TAG]=variable#adds the original to the df
			
def try_float_int(a,df,name):
	a=a.astype(float)
	try:
		
		if np.all(np.equal(np.mod(a, 1), 0)):
			a=a.astype(int)
	except:
		pass
	df[name]=a	

def convert_cat_to_int(a,df,name):
	print ("""Converting categorical variable %s to integers ...""" %(name,))
	q=np.unique(a)
	d=dict(zip(q, range(len(q))))
	df[name]=np.array([[d[k[0]]] for k in a],dtype=int)
	

def check_dateness(a,df,name,dateformat):
	n,k=a.shape
	a[a==np.array(None)]=''
	dts=np.unique(a)
	d=dict()
	lst=[]
	
	if 'datetime.date' in str(type(dts[0])):
		for dt in dts:
			d[dt]=(dt-date(1900,1,1)).days
			lst.append(d[dt])
	else:
		for dt in dts:
			d[dt]=(datetime.strptime(dt,dateformat)-datetime(1900,1,1)).days
			lst.append(d[dt])
	if np.max(lst)-np.min(lst)<3:#seconds
		if 'datetime.date' in str(type(dts[0])):
			for dt in dts:
				d[dt]=(dt-date(2000,1,1)).seconds
				lst.append(d[dt])
		else:
			for dt in dts:
				d[dt]=(datetime.strptime(dt,dateformat)-datetime(2000,1,1)).seconds
				lst.append(d[dt])
	df[name]=np.array([[d[k[0]]] for k in a])
	a=0




	
def get_best_sep(string,sep):
	"""Finds the separator that gives the longest array"""
	if not sep is None:
		return sep,string.split(sep)
	sep=''
	maxlen=0
	for i in [';',',',' ','\t']:
		b=head_split(string,i)
		if len(b)>maxlen:
			maxlen=len(b)
			sep=i
			c=b
	return sep,c,maxlen
				
def head_split(string,sep):
	a=string.split(sep)
	b=[]
	for j in a:
		if len(j)>0:
			b.append(j)	
	return b
			
def get_head_and_sep(fname,sep):
	f=open(fname,'r')
	head=f.readline().strip()
	r=[]
	sample_size=20
	for i in range(sample_size):
		r.append(f.read())	
	f.close()
	
	sep,h,n=get_best_sep(head,sep)
	for i in h:
		if is_number(i):
			raise RuntimeError("""The input file must contain a header row. No numbers are allowed in the header row""")
	
	for i in [sep,';',',','\t',' ']:#checks whether the separator is consistent
		err=False
		b=head_split(head, i)
		for j in r:
			if len(j.split(i))!=len(b):
				err=True
				break
			if err:
				h=b
				sep=i
			

	if sep is None:
		raise RuntimeError("Unable to find a suitable seperator for the input file. Check that your input file has identical number of columns in each row")
	return h,sep			
				
				
				
SQL_type_dict={0: float,
 1: int,
 2: int,
 3: float,
 4: float,
 5: float,
 6: float,
 8: int,
 9: int,
 16: int,
 246: int
 }


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains the argument class for the panel object

import stat_functions as stat
import numpy as np
import functions as fu
import loglikelihood as logl




class arguments:
	"""Sets initial arguments and stores static properties of the arguments"""
	def __init__(self,panel):
		p, q, d, k, m=panel.pqdkm
		self.categories=['beta','rho','lambda','gamma','psi','omega']
		if panel.z_active:
			self.categories+=['z']
		self.mu_removed=True
		if not self.mu_removed:
			self.categories+=['mu']
		
		self.panel=panel
		self.make_namevector(panel,p, q, k, m)
		self.set_init_args()
		self.position_defs()
		self.args_v=self.conv_to_vector(self.args_init)
		self.n_args=len(self.args_v)
		

	def initargs(self,p,d,q,m,k,panel):

		args=dict()
		args['beta']=np.zeros((panel.X.shape[2],1))
		args['omega']=np.zeros((panel.W.shape[2],1))
		args['rho']=np.zeros(p)
		args['lambda']=np.zeros(q)
		args['psi']=np.zeros(m)
		args['gamma']=np.zeros(k)
		args['omega'][0][0]=0
		args['mu']=np.array([])
		args['z']=np.array([])			
		if panel.m_zero and k>0:
			args['psi'][0]=1e-8
		
		if m>0 and panel.z_active:
			args['z']=np.array([1e-09])	

		if panel.N>1 and not self.mu_removed:
			args['mu']=np.array([0.0001])			
			

		return args

	def set_init_args(self,default=False):
		panel=self.panel
		p, q, d, k, m=panel.pqdkm
		
		args=self.initargs(p, d, q, m, k, panel)

		#de2=np.roll(e**2,1)-e**2
		#c=stat.correl(np.concatenate((np.roll(de2,1),de2),2),panel)[0,1]

		beta,e=stat.OLS(panel,panel.X,panel.Y,return_e=True)
		#beta=stat.OLS_simple(panel.input.Y,panel.input.X,True,False)
		self.init_e_st=e[panel.included]
		self.init_e_st=self.init_e_st/np.var(self.init_e_st)**0.5
		args['beta']=beta
		if panel.settings.group_fixed_random_eff.value==0 or panel.settings.variance_fixed_random_eff==0:
			args['omega'][0]=np.log(panel.var(e))

	
		self.args_start=fu.copy_array_dict(args)
		args_old=self.panel.input.args_archive.args
		if (not args_old is None) and (not default): 
			for name,ar_type in [['beta',False],
					  ['omega',False],
					  ['rho',True],
					  ['lambda',False],
					  ['psi',False],
					  ['gamma',True]]:
				self.insert_arg(name,args,ar_type)
			if panel.z_active:
				self.insert_arg('z',args)
			if not self.mu_removed:
				self.insert_arg('mu',args)
			
		self.args_init=args
		self.set_restricted_args(p, d, q, m, k,panel,e,beta)
		

	def set_restricted_args(self,p, d, q, m, k, panel,e,beta):
		self.args_restricted=self.initargs(p, d, q, m, k, panel)
		self.args_OLS=self.initargs(p, d, q, m, k, panel)		
		self.args_restricted['beta'][0][0]=np.mean(panel.Y)
		self.args_restricted['omega'][0][0]=np.log(np.var(panel.Y))
		self.args_OLS['beta']=beta
		self.args_OLS['omega'][0][0]=np.log((np.var(e*panel.included)*len(e[0])/np.sum(panel.included)))
	
	def create_null_ll(self):
		if not hasattr(self,'LL_OLS'):
			self.LL_OLS=logl.LL(self.args_OLS,self.panel).LL
			self.LL_null=logl.LL(self.args_restricted,self.panel).LL	
		
	def position_defs(self):
		"""Defines positions in vector argument"""

		self.positions=dict()
		self.positions_map=dict()#a dictionary of indicies containing the string name and sub-position of index within the category
		k=0
		for i in self.categories:
			n=len(self.args_init[i])
			rng=range(k,k+n)
			self.positions[i]=rng
			for j in rng:
				self.positions_map[j]=[0,i,j-k]#equation,category,position
			k+=n
	
	def conv_to_dict(self,args):
		"""Converts a vector argument args to a dictionary argument. If args is a dict, it is returned unchanged"""
		if type(args)==dict:
			return args
		if type(args)==list:
			args=np.array(args)			
		d=dict()
		k=0
		for i in self.categories:
			n=len(self.positions[i])
			rng=range(k,k+n)
			d[i]=args[rng]
			if i=='beta' or i=='omega':
				d[i]=d[i].reshape((n,1))
			k+=n
		return d


	def conv_to_vector(self,args):
		"""Converts a dict argument args to vector argument. if args is a vector, it is returned unchanged.\n
		If args=None, the vector of self.args_init is returned"""
		if type(args)==list or type(args)==np.ndarray:
			return np.array(args)
		v=np.array([])
		for i in self.categories:
			s=args[i]
			if type(s)==np.ndarray:
				s=s.flatten()
			v=np.concatenate((v,s))
		return v


	def make_namevector(self,panel,p, q, k, m):
		"""Creates a vector of the names of all regression varaibles, 
		including variables, ARIMA and GARCH terms. This defines the positions
		of the variables througout the estimation."""
		d=dict()
		names=panel.input.x_names[:]#copy variable names
		d['beta']=list(names)
		add_names(p,'rho%s    AR    p','rho',d,names)
		add_names(q,'lambda%s MA    q','lambda',d,names)
		add_names(k,'gamma%s  GARCH k','gamma',d,names)
		add_names(m,'psi%s    ARCH  m','psi',d,names)
		
		
		d['omega']=panel.input.W_names
		names.extend(panel.input.W_names)
		if m>0:
			if panel.N>1 and not self.mu_removed:
				d['mu']=['mu (var.ID eff.)']
				names.extend(d['mu'])
			if panel.z_active:
				d['z']=['z in h(e,z)']
				names.extend(d['z'])
			
		self.names_v=names
		self.names_d=d
		
	def insert_arg(self,argname,args,AR_type=False):
		arg=args[argname]
		arg_old=self.panel.input.args_archive.args[argname]
		names=self.names_d[argname]
		names_old=self.panel.input.args_archive.names_d[argname]
		n=min((len(arg),len(arg_old)))
		for i in names_old:
			if not i in names:
				return
		if n==0:
			return
		if len(arg.shape)==2:
			arg[:,0]=0
		else:
			arg[:]=0
		if not AR_type or len(arg_old)<=n:
			for i in range(len(names)):
				if names[i] in names_old:
					arg[i]=arg_old[names_old.index(names[i])]
		else:
			arg[:n-1]=arg_old[:n-1]
			arg[n-1]=np.sum(arg_old[n-1:])
	
			
def add_names(T,namsestr,category,d,names):
	a=[]
	for i in range(T):
		a.append(namsestr %(i,))
	names.extend(a)
	d[category]=a


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains classes used in the regression

import stat_functions as stat
import numpy as np
import time
import threading
import debug
import functions as fu
import calculus_functions as cf
import model_parser
import calculus
import copy
import loglikelihood as logl
from scipy import sparse as sp
import random_effects as re
import arguments

NON_NUMERIC_TAG='|~|'
min_AC=0.000001

#todo: split up panel in sub classes
def posdef(a,da):
	return list(range(a,a+da)),a+da

class panel:
	def __init__(self,dataframe,datainput,settings,pqdkm=None):

		self.input=datainput
		self.settings=settings
		self.dataframe=dataframe
		if not pqdkm is None:
			self.pqdkm=pqdkm
		else:
			self.pqdkm=settings.pqdkm.value
		self.initial_defs()
		self.arrayize()
		self.masking()
		self.lag_variables()
		self.final_defs()
		

	def initial_defs(self):
		if np.all(np.var(self.input.Y,0)==0):
			raise RuntimeError("No variation in Y")
		p,q,d,k,m=self.pqdkm
		self.max_lags=self.input.lag_obj.max_lags
		self.lost_obs=max((p,q,self.max_lags))+max((m,k,self.max_lags))+d#+3
		self.nW,self.n_beta=self.input.W.shape[1],self.input.X.shape[1]
		self.define_h_func()
		if self.input.IDs_name is None:
			self.settings.group_fixed_random_eff.value=0
		if self.input.timevar is None:
			self.settings.group_fixed_random_eff.value=0		
		self.m_zero = False
		if  m==0 and k>0:
			self.m_zero = True
			p,q,d,k,m=self.pqdkm
			self.pqdkm=p,q,d,k,1
			
	def masking(self):
		
		#"initial observations" mask: 
		self.a=np.array([self.date_counter<self.T_arr[i] for i in range(self.N)])# sets observations that shall be zero to zero by multiplying it with the arrayized variable
	
		#"after lost observations" masks: 
		self.T_i=np.sum(self.included,1).reshape((self.N,1,1))#number of observations for each i
		self.T_i=self.T_i+(self.T_i<=0)#ensures minimum of 1 observation in order to avoid division error. If there are no observations, averages will be zero in any case	
		self.N_t=np.sum(self.included,0).reshape((1,self.max_T,1))#number of observations for each t
		self.N_t=self.N_t+(self.N_t<=0)#ensures minimum of 1 observation in order to avoid division error. If there are no observations, averages will be zero in any case	
		self.group_var_wght=1-1/np.maximum(self.T_i-1,1)
		
	def final_defs(self):
		self.W_a=self.W*self.a
		self.tot_lost_obs=self.lost_obs*self.N
		self.NT=np.sum(self.included)
		self.NT_before_loss=self.NT+self.tot_lost_obs				
		self.number_of_RE_coef=self.N*(self.settings.group_fixed_random_eff.value>0)+self.n_dates*(self.settings.time_fixed_random_eff.value>0)
		self.number_of_RE_coef_in_variance=(self.N*(self.settings.group_fixed_random_eff.value>0)
											+self.n_dates*(self.settings.time_fixed_random_eff.value>0))*(self.settings.variance_fixed_random_eff.value>0)
		self.args=arguments.arguments(self)
		self.df=self.NT-self.args.n_args-self.number_of_RE_coef-self.number_of_RE_coef_in_variance
		a=0

	def lag_variables(self):
		T=self.max_T
		d=self.pqdkm[2]
		self.I=np.diag(np.ones(T))
		

		#differencing:
		if d==0:
			return
		L0=np.diag(np.ones(T-1),-1)
		Ld=(self.I-L0)
		for i in range(1,d):
			Ld=cf.dot(self.I-L0,Ld)		
		self.Y=cf.dot(Ld,self.Y)*self.a	
		self.X=cf.dot(Ld,self.X)*self.a
		if self.input.has_intercept:
			self.X[:,:,0]=1
		self.Y[:,:d]=0
		self.X[:,:d]=0	

	def params_ok(self,args):
		a=self.q_sel,self.p_sel,self.M_sel,self.K_sel
		for i in a:
			if len(i)>0:
				if np.any(np.abs(args[i])>0.999):
					return False
		return True


	def arrayize(self):
		"""Splits X and Y into an arry of equally sized matrixes rows equal to the largest for each IDs"""
		X, Y, W, IDs=self.input.X, self.input.Y, self.input.W, self.input.IDs
		timevar=self.input.timevar
		NT,k=X.shape
		if IDs is None:
			self.X=X.reshape((1,NT,k))
			self.Y=Y.reshape((1,NT,1))
			for i in [0,1]:
				if self.input.tobit_active[i]:
					self.tobit_I=self.input.tobit_I[i].reshape((1,NT,1))
			NTW,k=W.shape
			self.W=W.reshape((1,NT,k))
			self.time_map=None
			self.N=1
			self.max_T=NT
			self.T_arr=np.array([[NT]])
			self.date_counter=np.arange(self.max_T).reshape((self.max_T,1))
			self.included=np.array([(self.date_counter>=self.lost_obs)*(self.date_counter<self.T_arr[i]) for i in range(self.N)])
		else:
			sel,ix=np.unique(IDs,return_index=True)
			N=len(sel)
			sel=(IDs.T==sel.reshape((N,1)))
			T=np.sum(sel,1)
			self.max_T=np.max(T)
			idincl=T>self.lost_obs+self.settings.min_group_df.value
			self.X=arrayize(X, N,self.max_T,T, idincl,sel)
			self.Y=arrayize(Y, N,self.max_T,T, idincl,sel)
			self.tobit_I=[None,None]
			for i in [0,1]:
				self.tobit_I[i]=arrayize(self.input.tobit_I[i], N,self.max_T,T, idincl,sel,dtype=bool)
			self.W=arrayize(W, N,self.max_T,T, idincl,sel)
			self.N=np.sum(idincl)
			self.T_arr=T[idincl].reshape((self.N,1))
			self.date_counter=np.arange(self.max_T).reshape((self.max_T,1))
			self.included=np.array([(self.date_counter>=self.lost_obs)*(self.date_counter<self.T_arr[i]) for i in range(self.N)])
			self.get_time_map(timevar, self.N,T, idincl,sel)
			
	
			
			if np.sum(idincl)<len(idincl):
				idname=self.input.IDs_name[0]
				if idname + NON_NUMERIC_TAG in self.dataframe:
					id_orig=self.dataframe[idname + NON_NUMERIC_TAG]
					idremoved=id_orig[ix,0][idincl==False]
				else:
					idremoved=(self.dataframe[idname])[ix,0][idincl==False]
				s=fu.formatarray(idremoved,90,', ')
				print(f"The following {idname}s were removed because of insufficient observations:\n %s" %(s))
		self.zeros=np.zeros((self.N,self.max_T,1))
		self.ones=np.ones((self.N,self.max_T,1))

	
	def get_time_map(self,timevar, N,T_count, idincl,sel):
		if timevar is None:
			return None
		incl=self.included
		N,T,k=incl.shape
		unq,ix=np.unique(timevar,return_inverse=True)
		t=arrayize(np.array(ix).reshape((len(timevar),1)), 
				   N,self.max_T,T_count, idincl,sel,int)#maps N,T -> unique date
		grp_cnt=incl*np.arange(N).reshape(N,1,1)
		t_cnt=incl*np.arange(T).reshape(1,T,1)
		incl=incl[:,:,0]
		t=np.concatenate((t[incl],  grp_cnt[incl], t_cnt[incl]),1)
		a=np.argsort(t[:,0])
		t=t[a]#three columns: unique date index, group number, day sequence
		
	
		tid=t[:,0]#unique date
		t_map=[[] for i in range(np.max(tid)+1)]#all possible unique dates
		for i in range(len(tid)):
			t_map[tid[i]].append(t[i,1:])#appends group and day sequence
		t_map_tuple=[]
		tcnt=[]
		self.date_count_mtrx=np.zeros((N,T,1))
		for i in range(len(t_map)):
			a=np.array(t_map[i]).T#group and day sequence for unique date i
			if len(a):
				m=(tuple(a[0]),tuple(a[1]))#group and day sequence reference tuple
				n_t=len(a[0])#number of groups at this unique date
				t_map_tuple.append(m)	#group and day sequence reference tuple, for each unique date
				tcnt.append(n_t) #count of groups at each date
				self.date_count_mtrx[m]=n_t#timeseries matrix of the group count
				
		
		#A full random effects calculation is infeasible because of complexity and computing costs. 
		#A quazi random effects weighting is used. It  is more conservative than the full
		#RE weight theta=1-sd_pooled/(sd_pooled+sd_within/T)**0.5
		#If the weights are too generous, the RE adjustment may add in stead of reducing noise. 
		n=len(tcnt)
		self.n_dates=n
		self.date_count=np.array(tcnt).reshape(n,1,1)
		self.date_map=t_map_tuple
		
		
	def get_time_map2(self,timevar, N,T_count, idincl,sel):
		if timevar is None:
			return None
		unq,ix=np.unique(timevar,return_inverse=True)
		n_dates=len(unq)
		t=arrayize(np.array(ix).reshape((n_dates,1)), N,self.max_T,T_count, idincl,sel,int)#maps N,T -> unique date
		N,T,k=t.shape

		
		t_map=[[] for i in range(n_dates)]#all possible unique dates
		for i in range(len(tid)):
			t_map[tid[i]].append(t[i,1:])#appends group and day sequence
		t_map_tuple=[]
		tcnt=[]
		self.date_count_mtrx=np.zeros((N,T,1))
		for i in range(len(t_map)):
			a=np.array(t_map[i]).T#group and day sequence for unique date i
			if len(a):
				m=(tuple(a[0]),tuple(a[1]))#group and day sequence reference tuple
				n_t=len(a[0])#number of groups at this unique date
				t_map_tuple.append(m)	#group and day sequence reference tuple, for each unique date
				tcnt.append(n_t) #count of groups at each date
				self.date_count_mtrx[m]=n_t#timeseries matrix of the group count
				
		
		#A full random effects calculation is infeasible because of complexity and computing costs. 
		#A quazi random effects weighting is used. It  is more conservative than the full
		#RE weight theta=1-sd_pooled/(sd_pooled+sd_within/T)**0.5
		#If the weights are too generous, the RE adjustment may add in stead of reducing noise. 
		n=len(tcnt)
		self.n_dates=n
		self.date_count=np.array(tcnt).reshape(n,1,1)
		self.date_map=t_map_tuple
	
	
	def define_h_func(self):
		h_def="""
def h(e,z):
	e2			=	e**2+1e-5
	h_val		=	np.log(e2)	
	h_e_val		=	2*e/e2
	h_2e_val	=	2/e2-4*e**2/e2**2

	return h_val,h_e_val,h_2e_val,None,None,None
		"""	
		h_definition=self.settings.h_function.value
		if h_definition is None:
			h_definition=h_def
		d=dict()
		try:
			exec(h_definition,globals(),d)
			ret=d['h'](1,1)
			if len(ret)!=6:
				raise RuntimeError("""Your custom h-function must return exactly six arguments
				(h, dh/dx and ddh/dxdx for both e and z. the z return values can be set to None)""")
			self.h_def=h_definition
		except Exception as e:
			print('Something is wrong with your custom function, default is used:'+ str(e))
			exec(h_def,globals(),d)
			self.h_def=h_def
		
		self.z_active=True
		for i in ret[3:]:
			self.z_active=self.z_active and not (i is None)	
		if not self.settings.user_constraints.value is None:
			if not self.z_active and 'z' in self.settings.user_constraints.value:
				self.settings.user_constraints.value.pop('z')
			

		
		
	def mean(self,X,axis=None):
		dims=list(X.shape)
		dims[2:]=[1]*(len(dims)-2)
		#X=X*self.included.reshape(dims)
		if axis is None:
			return np.sum(X)/self.NT
		if axis==1:
			dims.pop(1)
			return np.sum(X,1)/self.T_i.reshape(dims)
		if axis==0:
			dims.pop(0)
			return np.sum(X,0)/self.N_t.reshape(dims)
		if axis==(0,1):
			return np.sum(np.sum(X,0),0)/self.NT
			
	def var(self,X,axis=None,k=1,mean=None,included=None):
		dims=list(X.shape)
		dims_m=np.array(X.shape)
		dims[2:]=[1]*(len(dims)-2)	
		if included is None:
			a=self.included
		else:
			a=included
		if mean is None:
			m=self.mean(X*a, axis)
		else:
			m=mean

		if axis==None:
			Xm=(X-m)*a.reshape(dims)
			return np.sum(Xm**2)/(self.NT-k)

		if axis==1:
			dims_m[1]=1
			m=m.reshape(dims_m)
			Xm=(X-m)*a.reshape(dims)	
			dims.pop(1)
			return np.sum((Xm)**2,1)/np.maximum(self.T_i-k,1).reshape(dims)
		if axis==0:
			dims_m[0]=1		
			m=m.reshape(dims_m)
			Xm=(X-m)*a.reshape(dims)
			dims.pop(0)
			return np.sum((Xm)**2,0)/np.maximum(self.N_t-k,1).reshape(dims)
		if axis==(0,1):
			dims_m[0:2]=1
			m=m.reshape(dims_m)
			Xm=(X-m)*a.reshape(dims)			
			return np.sum((Xm)**2,axis)/(self.NT-k)
	
def arrayize(X,N,max_T,T,idincl,sel,dtype=None):
	if X is None:
		return None
	NT,k=X.shape
	if dtype is None:
		Xarr=np.zeros((N,max_T,k))
	else:
		Xarr=np.zeros((N,max_T,k),dtype=dtype)
	T_used=[]
	k=0
	for i in range(len(sel)):
		if idincl[i]:
			Xarr[k,:T[i]]=X[sel[i]]
			k+=1
	Xarr=Xarr[:k]
	return Xarr




	



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import functions as fu

import loaddata
import numpy.lib.recfunctions as rfn



def parse_model(model_string):
	"""Parses model and returns the name of the Y variable and a list of the X variables as a tuple"""
	try:
		Y,X=model_string.split('~')
	except:
		raise RuntimeError("the 'model' argument should be a string on the form 'Y~X1 X2 X3' as separators for X-variables you can use either linefeed, space '+' or ','")
	Y=fu.clean(Y)
	X=fu.split_input(X)
	return Y,X


def test_dictionary(dataframe):
	n=0
	for i in dataframe.keys():
		if type(dataframe[i])==np.ndarray:
			if len(dataframe[i].shape)==1:
				dataframe[i]=dataframe[i].reshape((dataframe[i].shape[0],1))#converts to two dimensional
			if len(dataframe[i].shape)==2:
				if dataframe[i].shape[1]>1:
					raise RuntimeError("Variable %s has more than one column. This is not allowed." %(i,))		
			elif len(dataframe[i].shape)>2:
				raise RuntimeError("%s is a numpy ndarray with shape %s. Only one or two dimensional arrays are allowed" %(i,dataframe[i].shape))	
			if n==0:
				n=len(dataframe[i])
				s=i
			if len(dataframe[i])!=n:
				raise RuntimeError("Variables %s and %s have unequal number of observations. The number of observations must be identical for all variables. " %(i,s))	
			if n==0:
				raise RuntimeError("Variable %s has no observations. " %(i))	
	return n


def get_variables(input_class,dataframe,model_string,IDs_name,time_name,settings):
	print ("Analyzing variables ...")
	for i in dataframe:
		if type(dataframe[i])==np.ndarray:
			if len(dataframe[i])==0:
				raise RuntimeError("There are no valid observations")
			else:
				break
	sort(dataframe,time_name,IDs_name)
	
	input_class.timevar,input_class.time_name,void=check_var(dataframe,time_name,'time_name')
	input_class.IDs,input_class.IDs_name,void=check_var(dataframe,IDs_name,'ID_name')
	input_class.lag_obj=lag_object(input_class.IDs)
	dataframe['L']=input_class.lag_obj.lag#allowing for lag operator in model input
	input_class.W,input_class.W_names,void=check_var(dataframe,settings.heteroscedasticity_factors,'heteroscedasticity_factors',
                                       intercept_name='log variance constant',raise_error=False,intercept_variable=True)
	intercept_name=None
	if settings.add_intercept.value:
		intercept_name='Intercept'
	(input_class.X,input_class.x_names,
	 input_class.has_intercept,
	 input_class.Y,input_class.y_name)=parse_and_check(dataframe,model_string,intercept_name)
	
	tobit_set_threshold(input_class,settings)
	
		


def tobit_set_threshold(input_class,settings):
	"""Sets the tobit threshold"""
	tobit_limits=settings.tobit_limits.value
	Y=input_class.Y
	if tobit_limits is None:
		return
	if len(tobit_limits)!=2:
		print("Warning: The tobit_limits argument must have lenght 2, and be on the form [floor, ceiling]. None is used to indicate no active limit")
	if (not (tobit_limits[0] is None)) and (not( tobit_limits[1] is None)):
		if tobit_limits[0]>tobit_limits[1]:
			raise RuntimeError("floor>ceiling. The tobit_limits argument must have lenght 2, and be on the form [floor, ceiling]. None is used to indicate no active limit")
	g=[1,-1]
	input_class.tobit_active=[False, False]#lower/upper threshold
	input_class.tobit_I=[None,None]
	desc=['tobit_low','tobit_high']
	for i in [0,1]:
		input_class.tobit_active[i]=not (tobit_limits[i] is None)
		if input_class.tobit_active[i]:
			if np.sum(g[i]*Y<g[i]*tobit_limits[i]):
				print("Warning: there are observations of Y outside the non-cencored interval. These ")
			I=g[i]*Y<=g[i]*tobit_limits[i]
			Y[I]=tobit_limits[i]
			if np.var(Y)==0:
				raise RuntimeError("Your tobit limits are too restrictive. All observationss were cencored.")
			input_class.tobit_I[i]=I
			if np.sum(I)>0 and np.sum(I)<len(I):#avoiding singularity
				input_class.X=np.concatenate((input_class.X,I),1)
				input_class.x_names.append(desc[i])
		
def parse_and_check(dataframe,model_string,intercept_name):
	y_name,x_names=parse_model(model_string)
	X,x_names,has_intercept=check_var(dataframe,x_names,'x_names',intercept_name=intercept_name,raise_error=True,intercept_variable=True)
	Y,y_name,void=check_var(dataframe,y_name,'y_name',raise_error=True)	
	return X,x_names,has_intercept,Y,y_name

class lag_object:

	def __init__(self,IDs):
		self.IDs=IDs
		self.max_lags=0
		
	def lag(self,variable,lags=1):
		v=np.roll(variable, lags)
		idroll=np.roll(self.IDs,lags)
		keep=idroll==self.IDs
		self.max_lags=max((self.max_lags,lags))
		return v*keep
		

def sort(dataframe,time_name,IDs_name):
	"sorts the dataset"
	if (time_name is None) and (IDs_name is None):
		return
	elif (time_name is None) and (not IDs_name is None):
		g=dataframe[IDs_name]
		if np.var(g)==0:
			return
		srt=np.argsort(g,0).flatten()
	elif (not time_name is None) and (IDs_name is None):
		dt=dataframe[time_name]
		if np.var(g)==None:
			return
		srt=np.argsort(dt,0).flatten()	
	else:
		dt=dataframe[time_name]
		vdt=np.var(dt)
		dt=dt.astype(dtype=[('date',type(dt[0][0]))])
		g=dataframe[IDs_name]
		vg=np.var(g)
		g=g.astype(dtype=[('IDs',type(g[0][0]))])
		if vdt==0 and vg==0:
			return
		if vdt==0 and vg>0:
			srt=np.argsort(g,0).flatten()
		elif vdt>0 and vg==0:
			srt=np.argsort(dt,0).flatten()
		else:
			s=rfn.merge_arrays((dt,g))
			srt=np.argsort(s,0,order=['IDs','date']).flatten()
	for i in dataframe:
		if type(dataframe[i])==np.ndarray:
			dataframe[i]=dataframe[i][srt]

	

def modify_dataframe(dataframe,transforms=None,filters=None):
	n=test_dictionary(dataframe)
	if 'ones' in dataframe.keys():
		print ("Warning: variable 'ones' is replaced by a constant vector of ones. Do not give any variable this name if you want to avoid this.")
	dataframe['ones']=np.ones((n,1))
	if not transforms is None:
		exec(transforms,globals(),dataframe)	
	if not filters is None:
		n=filter_data(filters, dataframe)
		if not transforms is None:
			exec(transforms,globals(),dataframe)
	for i in list(dataframe.keys()):
		if callable(dataframe[i]):
			dataframe.pop(i)		
	print ("Checking and parsing variables ...")

def check_var(dataframe,names,arg_name,intercept_name=None,raise_error=False,intercept_variable=False):
	if names is None:
		if intercept_name is None:
			return None,None,None
		else:
			names=[]
	has_const=False
	if type(names)==str:
		names=[names]
	if type(names)!=list:
		raise RuntimeError("The %s argument must be a string or a list of strings" %(arg_name,))
	check_and_add_variables(names, dataframe, arg_name)
	if len(names)>0:
		has_const=np.all(dataframe[names[0]]==1)	
	if not intercept_name is None:
		if not has_const:
			names=[intercept_name]+names
		else:
			names[0]=intercept_name
		dataframe[intercept_name]=dataframe['ones']
		has_const=True
	X=[]
	for i in names:
		X.append(dataframe[i])
	X=np.concatenate(X,1)
	X,names=remove(X,names, dataframe, raise_error,has_const)
	return X,names,has_const

	
def remove(X,names,dataframe,raise_error,has_const):
	"""Removes constants variables at position 1 or higher, and any zero variable. You shold set raise_error=True for vital variables (X and Y)"""
	
	keep=np.var(X,0)!=0
	keep[0]=has_const or keep[0]#allways keep first variable unless all observations are aproximately zero
	sumtrash=np.sum(keep==0)
	if sumtrash>0:
		remvd=','.join(np.array(names)[keep==False])
		if sumtrash==1:
			remvd="Warning: The variable %s was removed because it was requested removed or constant" %(remvd,)
		else:
			remvd="Warning: The variables %s were removed because they were requested removed or constant" %(remvd,)
		if raise_error and len(X[0])<=sumtrash:
			raise RuntimeError(remvd+'. Aborting since there are no more variables to run with.')
		else:
			a=0
			#print(remvd)
		if len(X[0])<=sumtrash:
			return None,None
	X=X[:,keep]
	names=list(np.array(names)[keep])
	return X,names

	

	
def check_and_add_variables(names,dataframe,arg_name):
	if names is None:
		return
	elif type(names)==str:
		names=[names]
	for name in names:
		if not name in dataframe:
			if (name.lower() in ['constant','ones','intercept','one','alpha']):
				dataframe[name]=dataframe['ones']
			else:
				try:
					var=eval(name,globals(),dataframe)
				except KeyError as e:
					raise RuntimeError("Variable %s is requested in %s, but it does not exist in the dataframe" %(name,arg_name))
				dataframe[name]=var
			
			

def filter_data(filters,dataframe,copy=True):
	"""Filters the dataframe based on setting in the string Filters"""
	if filters is None:
		return
	if not ' and ' in filters:
		fltrs=filters.split()
	else:
		fltrs=fu.clean(filters,' and ')
	n=len(dataframe[list(dataframe.keys())[0]])
	v=np.ones(n,dtype=bool)
	for f in fltrs:
		r=eval(f,globals(),dataframe)
		r.resize(n)
		print ('Removing %s observations due to filter %s' %(np.sum(r==0),f))
		v*=r
	if copy:
		d=dict()
	else:
		d=dataframe
	for i in dataframe:
		if type(dataframe[i])==np.ndarray:
			if len(dataframe[i])==n:
				d[i]=dataframe[i][v]
				k=len(d[i])
	print ('Removed %s of %s observations - %s observations remaining' %(n-k,n,k))
	return d,k



			
		
def get_data_and_model(X,Y,W=None,IDs=None,x_names=None,y_name=None,heteroscedasticity_factors=None,IDs_name=None,filters=None,transforms=None):
	"""Complies X and Y (and if supplied also W and IDs) to a dataframe, and returns:\n
	- a dictionary with all variables """
	dataframe=dict()
	x_names=add_var_to_dict(dataframe, X, x_names, 'X')
	y_name=add_var_to_dict(dataframe, Y, y_name, 'Y')
	heteroscedasticity_factors=add_var_to_dict(dataframe, W, heteroscedasticity_factors, 'W')
	IDs_name=add_var_to_dict(dataframe, IDs, IDs_name, 'IDs')
	
	model_string="%s ~ %s" %(y_name[0], '+'.join(x_names))
	if not heteroscedasticity_factors is None:
		heteroscedasticity_factors=','.join(heteroscedasticity_factors)	
	modify_dataframe(dataframe,filters,transforms)
	return dataframe, model_string, heteroscedasticity_factors, IDs_name
		


def add_var_to_dict(dataframe,V,v_names,arg_name):
	if V is None:
		if v_names is None:
			return None
		else:
			if type(v_names)==str:
				v_names=fu.clean(v_names,",")	
			return v_names
	if type(V)==dict:
		v_names=[]
		for i in V.keys():
			dim_check(V[i],arg_name,dict_check=True)
			dataframe[i]=V[i]
			v_names.append(i)
	elif type(V)==np.ndarray:
		v_names=dim_check(V,arg_name,v_names)
		for i in range(len(v_names)):
			dataframe[v_names[i]]=V[:,i:i+1]
	return v_names
			
			
			
		
def dim_check(V,arg_name,v_names=None,dict_check=False):
	if len(V.shape)!=2:
		raise RuntimeError ("%s is not two dimensional. Only two dimensional numpy_arrays are allowed" %(arg_name))	
	if dict_check:
		return
	N,k=V.shape
	if v_names is None:
		v_names=[arg_name+str(i) for i in range(k)]
	else:
		if type(v_names)==str:
			v_names=fu.clean(v_names,",")
		if len(v_names)!=k:
			raise RuntimeError ("A name list is supplied for %s, but its lenght does not match the number of columns in %s." %(arg_name,arg_name))	
	return v_names


def check_sign(results,stats,category,oldp,fixed,maxp=1000,sign_level=0.05):
	"Checks whether the higest order of category is significant. If it is not, lim is set to True"
	sign=stats.output.tsign[results.panel.args.positions[category]]
	if len(sign)==0:
		return 1,fixed
	members=results.panel.args.positions[category]
	a=list(results.constraints.fixed)
	b=list(results.constraints.associates)
	c=np.unique(a+b)
	n=len(sign)
	sign=sign[::-1]
	for i in range(n):
		constrained=np.any([j in c for j in members])
		if np.all(sign[i:])<sign_level and not constrained:
			if n-i<oldp or n-i+1>maxp:
				return n-i,True
			else:
				return n-i+1,False
	return 1,fixed




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from gui import gui_charts
from gui import gui_scrolltext
			
class output_tab(tk.Frame):
	def __init__(self,window,exe_tab):
		main_tabs=window.main_tabs
		name=main_tabs.gen_name('regression')
		tk.Frame.__init__(self,main_tabs)
		
		self.widget=tk.Frame(self)
		self.widget.columnconfigure(0,weight=1)
		self.widget.columnconfigure(1)
		self.widget.rowconfigure(0,weight=1)

		
		self.box= gui_scrolltext.ScrollText(self.widget,format_text=False,name='regression',window=window,)
		self.charts = gui_charts.process_charts(window,self.widget)
		self.progress_bar=bar(self,exe_tab)
		
		self.progress_bar.grid(row=2, sticky=tk.EW)
		

		self.tab=main_tabs._tabs.add(self,name=name,top_text='Regression output',top_color='#fcf3d9',)
		self.tab.exe_tab=exe_tab
		self.tab.widget=self.box
	
		self.widget.grid(row=1,column=0,sticky=tk.NSEW)
		
		self.box.grid(column=0,row=0,sticky=tk.NSEW)
		self.charts.grid(column=1,row=0,sticky=tk.NS)			
		
		
		main_tabs.select(self)
		main_tabs.insert('end',main_tabs.add_tab)			
		
	

class bar(tk.Frame):
	def __init__(self,master,exe_tab):
		tk.Frame.__init__(self,master,background='white',height=25)
		self.tab=master
		self.suffix=''
		self.exe_tab=exe_tab
		self.text=tk.StringVar(self)
		self.text_lbl=tk.Label(self,textvariable=self.text,background='white')
		self.progress=tk.Frame(self,background='#9cff9d',height=5,width=0)
		self.text_lbl.grid(row=0,column=0,sticky=tk.W)
		self.progress.grid(row=1,column=0,sticky=tk.W)
		
	def set_progress(self,percent,text):
		total_width=self.winfo_width()
		self.progress.config(width=int(total_width*percent))
		self.progress.grid()
		if len(self.suffix):
			text=self.suffix + ' - ' + text
		self.text.set(text)
		if not self.exe_tab is None:
			return self.exe_tab.isrunning
		else:
			return True


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import tkinter as tk
from tkinter import ttk
from gui import gui_scrolltext

import os
from tkinter import filedialog
import types
import numpy as np
from multiprocessing import pool
import paneltime as pt
import traceback
from matplotlib import pyplot  as plt



ADD_EDITOR_NAME='...'
DEFAULT_GREY='#e6e6e6'

class main_tabs(ttk.Notebook):
	def __init__(self,window):
		self.win=window
		ttk.Notebook.__init__(self,window.frm_left) 
		self.bind("<<NotebookTabChanged>>", self.main_tab_pressed)
		self.last_active_tab=None
		self.grid(column=0,row=0,sticky=tk.NSEW)  # Pack to make visible	
		self._tabs=tabs(self)
		self.add_tab = tk.Frame(self)
		self._tabs.add(self.add_tab,None,ADD_EDITOR_NAME)
		
		self.isdeleting=False				
		
	def recreate_tabs(self):
		if self.win.data.get('editor_data') is None:
			return
		n=0
		for i in self.win.data.get('editor_data'):
			try:
				name,text, top_text, top_color,attached_to,path=self.win.data.get('editor_data')[i]
				if not (name==ADD_EDITOR_NAME or (name[:6]=='script' and text=='')):
					self.add_editor(name,text,top_text=top_text,top_color=top_color,attached_to=attached_to,path=path).frame.focus()	
					n+=1
			except:
				print(f'Could not add tab {i}')
		if n==0:
			self.add_editor('script').frame.focus()		
			
	def main_tab_pressed(self,event):	
		s=self.current_editor(0)
		if s==ADD_EDITOR_NAME:
			tab=[self.add_editor()]
		else:
			self._tabs.selection_change(self.select())
			
	def insert_current_editor(self,chars):
		tb=self.current_editor(2)
		tb.write(chars)
		tb.text_box.focus()
		
	def current_editor(self,return_obj=1):
		selection = self.select()
		if return_obj==1:
			return str(self._tabs[selection].frame)#returns tab object
		elif return_obj==2:
			return self._tabs[selection].widget#returns widget
		return self.tab(selection, "text")

		
	def selected_tab_text(self):
		tb=self.current_editor(2)
		text=tb.get_all()	
		return text
	
	def gen_name(self,core):
		if core is None:
			core='script'
		if 	not core in self._tabs.names:
			return core
		i=0
		n=list(core.split('.'))
		if len(n)==1:
			while True:
				i+=1
				name=f"{core} {i}"
				if 	not name in self._tabs.names:
					return name	
		else:
			while True:
				i+=1
				name=f"{'.'.join(n[:-1])} {i}.{n[-1]}"
				if 	not name in self._tabs.names:
					return name				
			
	def add_editor(self,name=None,text=None,format_text=True,top_text='',top_color=DEFAULT_GREY,dataset=None,attached_to=None,path=None):
		name=self.gen_name(name)
		frame=tk.Frame(self)
		widget= gui_scrolltext.ScrollText(frame,text=text,format_text=format_text,name=name,window=self.win)
		tab=self._tabs.add(frame,widget,name,top_text,top_color,dataset,path,attached_to)
		frame.tab=tab
		self.select(frame)
		self.insert('end',self.add_tab)	
		self.remove_redundant_script(name)
		return tab
		
	def remove_redundant_script(self,name):
		if name[:6]=='script':
			return
		poptabs=[]
		for i in self._tabs:
			try:
				if self._tabs[i].name[:6]=='script' and self._tabs[i].widget.get_all().replace('\n','')=='':
					poptabs.append(i)
			except:
				pass
		for i in poptabs:
			self._tabs.pop(i)
'add_stat_tab'			
		
		
class tab:
	def __init__(self,tabs,frame,widget,name,top_text='',top_color=DEFAULT_GREY,dataset=None,path=None,attached_to=None):
		self.frame = frame
		self.position=tabs.count
		self.isrunning=False
		self.tabs=tabs
		self.name=name
		self.path=path
		self.attached_to=attached_to
		self.dataset=dataset		
		self.top_color=top_color
		self.frame.rowconfigure(0)
		self.frame.rowconfigure(1, weight=1)
		self.frame.columnconfigure(0, weight=1)	
		self.notebook=tabs.notebook
		self.locals=dict()
		self.globals={'window':self.notebook.win,'tab':self,'data':None}		
		self.add_buttons(name,top_color,attached_to,top_text)
		self.notebook.add(self.frame,text=name)
		self.widget = widget	
		self.locals=dict()
		if not widget is None:
			widget.grid(row=1, column=0,sticky=tk.NSEW)

		
	def add_buttons(self,name,top_color,attached_to,top_text):
		size=22
		self.button_frame=tk.Frame(self.frame,height=size,background=top_color)
		self.button_frame_L=tk.Frame(self.button_frame,height=size,background=top_color)
		self.button_frame_R=tk.Frame(self.button_frame,height=size,background=top_color)
		self.top_text=tk.StringVar(value=top_text,master=self.button_frame)
		self.button_frame_label_M=tk.Label(self.button_frame,textvariable=self.top_text,background=top_color)

		self.display_name=tk.StringVar(self.button_frame)
		self.display_name.set(name)
		self.name_box=tk.Entry(self.button_frame_L,width=50,textvariable=self.display_name)
		self.name_box.bind('<KeyRelease>', self.tab_name_edited)
		self.button_img=dict()		
		self.button_img['save']= tk.PhotoImage(file =  os.path.join(os.path.dirname(__file__),'img\\save.png'),master=self.button_frame)
		self.button_save=tk.Button(self.button_frame_L, image = self.button_img['save'],command=self.save, 
								   highlightthickness=0,bd=0,height=size, anchor=tk.E,background=top_color)
		
		self.button_img['run']= tk.PhotoImage(file =  os.path.join(os.path.dirname(__file__),'img\\run.png'),master=self.button_frame)
		self.button_run=tk.Button(self.button_frame_L, image = self.button_img['run'],command=self.run, 
								   highlightthickness=0,bd=0,height=size, anchor=tk.E,background=top_color)		
		
		self.button_img['stop']= tk.PhotoImage(file =  os.path.join(os.path.dirname(__file__),'img\\stop.png'),master=self.button_frame)
		self.button_stop=tk.Button(self.button_frame_L, image = self.button_img['stop'],command=self.stop, 
								   highlightthickness=0,bd=0,height=size, anchor=tk.E,background=top_color)				
		
		self.attach_text=tk.Label(self.button_frame_R,text="attached to",background=top_color)
		self.add_attached_box(self.button_frame_R,attached_to)
		
		self.button_img['delete']= tk.PhotoImage(file =  os.path.join(os.path.dirname(__file__),'img/delete_small.png'),master=self.button_frame)
		self.button_delete=tk.Button(self.button_frame_R, image = self.button_img['delete'],command=self.delete, 
								   highlightthickness=0,bd=0,height=15, anchor=tk.E,background=top_color)
		
		
		self.button_frame.grid(row=0,sticky=tk.EW)
		self.button_frame.columnconfigure(0,weight=1)
		self.button_frame.columnconfigure(1,weight=5)
		self.button_frame.columnconfigure(2,weight=1)
		self.button_frame_L.grid(row=0,column=0,sticky=tk.W)
		self.button_frame_label_M.grid(row=0,column=1,sticky=tk.EW)
		self.button_frame_R.grid(row=0,column=2,sticky=tk.E)
		
		self.name_box.grid(row=0,column=0)
		self.button_save.grid(row=0,column=1)
		self.button_run.grid(row=0,column=2)
		self.button_stop.grid(row=0,column=3)
		
		self.button_delete.grid(row=0,column=2)
		if not self.attached_to is None:
			self.attached_to.grid(row=0,column=1)
			self.attach_text.grid(row=0,column=0)
		
	def stop(self):
		self.tabs.notebook.win.mc.master.quit()
		self.tabs.notebook.win.mc=None
		if hasattr(self,'exe_tab'):
			self.exe_tab.isrunning=False
			self.exe_tab.pool.close()
		else:
			self.isrunning=False
			self.pool.close()	
		
	
	def run(self):
		win=self.notebook.win
		win.data.save()
		self.pool = pool.ThreadPool(processes=1)
		text=self.widget.get_all()
		self.globals['exe_tab']=self
		self.isrunning=True
		self.process=self.pool.apply_async(self.exec, (text,))

	def exec(self,source):
		try:
			exec(source,self.globals,self.locals)
		except Exception as e:
			print("""
The following error occured in you script:
------------------------------------------
""")
			traceback.print_exc()
	
	def run_enable(self,event):
		pass
		
	def add_attached_box(self,master,attached_to):
		if self.name==ADD_EDITOR_NAME:
			return		
		self.attached_to_text=tk.StringVar(master)
		self.attached_to = ttk.Combobox(master,textvariable=self.attached_to_text)
		self.attached_to.bind("<<ComboboxSelected>>",self.attachment_change)
		self.attached_to.bind("<Button-1>",self.attachment_add_items)
		self.attachment_add_items(attached_to=attached_to)
		
	
	def attachment_add_items(self,event=None,attached_to=None):
		if self.name==ADD_EDITOR_NAME:
			return
		if hasattr(self.notebook.win, 'right_tabs'):
			datasets=self.notebook.win.right_tabs.data_tree.datasets
		else:
			if hasattr(self,'datasets'):
				datasets=self.dataset.datasets
			else:
				return
		self.globals['data']=datasets
		attach_list=list(datasets.keys())+['']
		attach_list.sort()
		self.attached_to.config(values=attach_list)
		if not self.dataset is None:
			self.attached_to_text.set(self.dataset.name)
		elif not attached_to is None:
			self.attached_to_text.set(attached_to)
		else:
			self.attached_to_text.set('')
	
	def attachment_change(self,event):
		name=self.attached_to.get()
		if name=='':
			return
		self.dataset=self.notebook.win.right_tabs.data_tree.datasets[name]
		if self.top_text.get()=="Data editor":
			self.dataset.edit_editor=str(self.frame)
		elif self.top_text.get()=="Import script":
			self.dataset.script_editor=str(self.frame)
		elif self.top_text.get()=="Model execution editor":
			self.dataset.exe_editor=str(self.frame)
		
		
	def tab_name_edited(self,event=None):
		self.notebook.tab(self.frame,text=self.display_name.get())
		
		
	def edit_data_set(self):
		if self.top_text.get()!="Data editor":
			return
		if self.dataset is None:
			return
		d=self.dataset
		script=self.widget.get_all()
		exec(script,d,d)
		for i in list(d.keys()):
			if not type(d[i])==np.ndarray:
				d.pop(i)
		self.dataset.datasets.make_tree(self.dataset.name,self.notebook.win.right_tabs.data_tree)
		self.widget.focus()
		
	def save(self):
		p=self.notebook.win.data['current path']
		filename = filedialog.asksaveasfilename(initialdir=p,title="Save",
			filetypes = (("text", "*.txt"),("Rich Text Format", "*.rtf"),("Python file", "*.py")), defaultextension=True)
		if not filename: 
			return
		self.path=filename
		p,f=os.path.split(filename)
		self.notebook.win.data['current path']=p
		txt=self.widget.get_all()
		file = open(filename,'w')
		file.write(txt)
		file.close()	
	
	def delete(self):
		self.notebook.isdeleting=True
		self.tabs.pop(self.frame)
	
class tabs(dict):
	def __init__(self,notebook):
		dict.__init__(self)
		self.count=0
		self.names=dict()
		self.notebook=notebook
		self.sel_list=[]
		self.subplot=plt.subplots(1,figsize=(4,2.5),dpi=75)
		self.print_subplot=plt.subplots(1,figsize=(6,3.25),dpi=200)				
	
	def add(self,frame,widget=None,name='',top_text='',top_color=DEFAULT_GREY,dataset=None,path=None,attached_to=None):
		
		tab_item=tab(self,frame, widget, name,top_text,top_color,dataset,path,attached_to=attached_to)
		dict.__setitem__(self,str(frame), tab_item)
		self.names[name]=self[frame]
		self.count+=1
		self.sel_list.append(str(frame))
		return self[frame]
	
	def save_all_in_temp(self):
		win=self.notebook.win
		win.data['editor_data']=dict()
		for i in self:
			t=self[i]
			try:
				text=t.widget.get('1.0',tk.END)
			except:
				text=''
			if not t.name==ADD_EDITOR_NAME:
				win.data['editor_data'][i]=(t.name,text,t.top_text.get(),t.top_color,t.attached_to.get(),t.path)
	
	def __getitem__(self,k):
		k=str(k)
		ret=dict.__getitem__(self,k)
		return ret
		
	def __setitem__(self,k,v):
		raise RuntimeError("Can't assign values to the tabs object. Use the add method to create a new tab")
		
	def pop(self, k, d=None):
		if self[k].name==ADD_EDITOR_NAME:
			return
		add_new=False
		if self.count<=2:
			add_new=True
		self.select_valid_tab(k)		
		k=str(k)
		self.sel_list.pop(self.sel_list.index(k))
		try:
			self.names.pop(self[k].name)
		except:
			print(f"Could not delete {self[k].name} from {self.names}")
		self.notebook.forget(k)
		dict.pop(self,k, d)
		self.erase_invalid_attachments()
		self.count-=1
		if add_new:
			self.notebook.add_editor('script')
			self.count+=1
			
	def erase_invalid_attachments(self):
		datasets=self.notebook.win.right_tabs.data_tree.datasets
		for name in datasets:
			self.compare_attachment(datasets[name], 'edit_editor')
			self.compare_attachment(datasets[name], 'exe_editor')
			self.compare_attachment(datasets[name], 'script_editor')
					
	def compare_attachment(self,dataset,attribute):
		f=dataset.__dict__[attribute]
		found=False
		for i in self:
			if str(self[i])==f:
				found=True
				break
		if not found:
			dataset.__dict__[attribute]=None
		
		
		
		
	def selection_change(self,selection):
		i=self.sel_list.index(selection)
		self.sel_list.append(self.sel_list.pop(i))

	def select_valid_tab(self,current_tab):
		sel_list=self.sel_list
		for i in range(len(sel_list)):
			f=self[sel_list[-1-i]]
			if f.frame!=current_tab and f.name!='...':
				self.notebook.select(sel_list[-1-i])
				break
						
	

		


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import tkinter as tk
from tkinter import ttk
from tkinter import ttk
from multiprocessing import pool

import numpy as np
import stat_functions as stat
from scipy import stats as scstats
from gui import gui_functions as guif



class process_charts(ttk.Frame):
	def __init__(self,window,master):
		style = ttk.Style()
		style.configure("TFrame", background='white')		
		ttk.Frame.__init__(self,master,style='new.TFrame')
		self.window=window
		self.ll=None	
		self.initialized=False
		self.subplot=window.main_tabs._tabs.subplot
		self.print_subplot=window.main_tabs._tabs.print_subplot
		
	def add_content(self):
		self.n_charts=3
		self.columnconfigure(0,weight=1)
		for i in range(self.n_charts+1):
			self.rowconfigure(i,weight=1)
			

		tk.Label(self,text='Charts on normalized residuals:',bg='white',font='Tahoma 10 bold').grid(row=0,column=0)			

		self.charts=[]
		for i in range(self.n_charts):
			frm=tk.Frame(self,background='white')
			frm.rowconfigure(0,weight=1)
			frm.rowconfigure(1)
			frm.columnconfigure(0,weight=1)
			self.charts.append(tk.Label(frm,background='white'))
			self.charts[i].grid(row=0,column=0)	
			guif.setbutton(frm, 'Save image', lambda: self.save(self.n_charts-i-1),bg='white').grid(row=1,column=0)
			frm.grid(row=i+1)
		
	def save(self,i):
		if not hasattr(self.charts[i],'graph_file') or not hasattr(self,'panel'):
			print('No graphics displayed yet')
			return
		name=self.charts[i].name
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile=f"{name}.jpg")		
		if f is None:
			return
		flst=[
			self.histogram,
			self.correlogram,
			self.correlogram_variance,
		]
		flst[i](self.ll,print_subplot,f)
		f.close()
		
	def initialize(self,panel):
		if not self.initialized:
			self.panel=panel
			self.add_content()
			self.initialized=True		
		
	def plot(self,ll):
		self.ll=ll
		self.histogram(ll,self.subplot)
		self.correlogram(ll,self.subplot)
		self.correlogram_variance(ll,self.subplot)	
		
		
	def histogram(self,ll,subplot,f=None):
		fgr,axs=subplot
		e=ll.e_st_centered[self.panel.included]
		N=e.shape[0]
		e=e.reshape((N,1))
		
		grid_range=4
		grid_step=0.05	
		h,grid=histogram(e,grid_range,grid_step)
		norm=scstats.norm.pdf(grid)*grid_step	
		
		axs.bar(grid,h,color='grey', width=0.025,label='histogram')
		axs.plot(grid,norm,'green',label='normal distribution')
		axs.legend(prop={'size': 6})
		name='Histogram - frequency'
		axs.set_title(name)
		if f is None:
			guif.display(self.panel, self.charts[0],name,0,subplot)
		else:
			guif.save(subplot,f)

	def correlogram(self,ll,subplot,f=None):
		fgr,axs=subplot
		lags=20
		rho=stat.correlogram(self.panel, ll.e_st_centered,lags)
		x=np.arange(lags+1)
		axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
		name='Correlogram - residuals'
		axs.set_title(name)
		if f is None:
			guif.display(self.panel, self.charts[1],name,1,subplot)
		else:
			guif.save(subplot,f)
		
	def correlogram_variance(self,ll,subplot,f=None):
		fgr,axs=subplot
		lags=20
		e2=ll.e_st_centered**2
		e2=(e2-self.panel.mean(e2))*self.panel.included
		rho=stat.correlogram(self.panel, e2,lags)
		x=np.arange(lags+1)
		axs.bar(x,rho,color='grey', width=0.5,label='correlogram')
		name='Correlogram - squared residuals'
		axs.set_title(name)
		if f is None:
			guif.display(self.panel, self.charts[2],name,2,subplot)
		else:
			guif.save(subplot,f)
	
def histogram(x,grid_range,grid_step):
	N,k=x.shape
	grid_n=int(2*grid_range/grid_step)
	grid=np.array([i*grid_step-grid_range for i in range(grid_n)]).reshape((1,grid_n))
	ones=np.ones((N,1))
	x_u=np.concatenate((ones,x>=grid),1)
	x_l=np.concatenate((x<grid,ones),1)
	grid=np.concatenate((grid.flatten(),[grid[0,-1]+grid_step]))
	histogram=np.sum((x_u*x_l),0)
	if int(np.sum(histogram))!=N:
		raise RuntimeError('Error in histogram calculation')
	return histogram/N,grid

	
class scatter_charts(tk.Toplevel):
	def __init__(self, master,panel,X,Y,iconpath,height=400,width=1000):
		tk.Toplevel.__init__(self, master,height=400,width=1000,)
		
		self.panel=panel
		self.X=X
		self.Y=Y
		self.title('Scatter charts')
		self.geometry('%sx%s' %(width,height))
		self.iconbitmap(iconpath)
		self.rowconfigure(0,weight=1)
		self.columnconfigure(0,weight=1)			
		
		self.n_cols=3
		self.col_height=250
		self.n_plots=self.panel.input.X.shape[1]
		self.n_rows=int(self.n_plots/self.n_cols)
		self.n_rows+=(self.n_rows*self.n_cols<self.n_plots)
		
		self.yscrollbar = tk.Scrollbar(self)
			
		self.main_frame=tk.Frame(self)
		self.main_frame.grid(row=0,column=0)			
		
		self.btn_saveall = tk.Button(self.main_frame, text='Save all', command=self.saveall,bg='white')
		self.btn_saveall.grid(row=1,column=0)		
		
		self.canvas=tk.Canvas(self.main_frame,yscrollcommand = self.yscrollbar.set,scrollregion=(0,0,width,self.n_rows*self.col_height),height=height,width=width)
					

		
		
		self.wdgt_frame=tk.Frame(self.canvas,width=width,height=self.n_rows*self.col_height)
			
		
		self.yscrollbar.grid(row=0,column=1,sticky='ns')
		self.wdgt_frame.grid(row=0,column=0,sticky=tk.NSEW)
		self.canvas.grid(row=0,column=0,sticky=tk.NSEW)	
		self.plot_all()	
		
		
		
			
		
		self.canvas.create_window(0,0,window=self.wdgt_frame,anchor='nw')
		self.yscrollbar.config(command = self.canvas.yview)	

		self.transient(master) #set to be on top of the main window
		self.grab_set() #hijack all commands from the master (clicks on the main window are ignored)
		#master.wait_window(self) #pause anything on the main window until this one closes (optional)


	def plot_all(self):
		self.charts=dict()
		for i in range(self.n_rows):
			self.wdgt_frame.rowconfigure(i,weight=1)	
		for i in range(self.n_cols):
			self.wdgt_frame.columnconfigure(i,weight=1)			
		for row in range(self.n_rows):
			for col in range(self.n_cols):
				i=row*self.n_cols+col
				if i>=self.n_plots:
					break
				self.charts[(row,col)]=self.plot(i,row,col,bgframe=self.wdgt_frame)
		for i in self.charts:			
			self.charts[i].grid(row=i[0],column=i[1])	
			
				
		
				
						
		
	def plot(self,i,row,col,subplot=None,f=None,bgframe=None):
		if subplot is None:
			subplot=self.subplot
		fgr,axs=subplot
		x_name=self.panel.input.x_names[i]
		y_name=self.panel.input.y_name[0]
		x=self.X[:,i]
		y=self.Y[:,0]
		
		axs.scatter(x,y, alpha=.1, s=10)
		axs.yaxis.label.set_text(y_name)
		axs.yaxis.label.set_text(x_name)
		name=f'{y_name} - {x_name}'
		axs.set_title(name)	
		
		if f is None:
			w=int(fgr.get_figwidth()*fgr.get_dpi())
			h=int(fgr.get_figheight()*fgr.get_dpi())
			chart=tk.Label(bgframe,width=w,height=h)
			guif.display(self.panel, chart,name,i,subplot,self.on_scatter_click)
			return chart
		
		else:
		
			guif.save(subplot,f)
					
	def on_scatter_click(self,event):
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile=f"{event.widget.name}.jpg")	
		frame=tk.Label(self.wdgt_frame)	
		self.plot(event.widget.i,subplot=self.print_subplot, f=f)
		frame.grid(row=0,column=0)		
		f.close()		
		
	def saveall(self):
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile="paneltime_scatter_plots.jpg")	
		fname=f.name
		f.close()
		for i in self.charts:
			f=open(gui.fix_fname(fname,i.name))
			self.plot(i.i, self.print_subplot,f)
			f.close()
			
	def on_closing(self):
		self.withdraw()
		
	def save(self,i):
		f = tk.filedialog.asksaveasfile(mode='bw', defaultextension=".jpg",initialfile=f"paneltime_scatter_plot_{i}.jpg")		
		self.plot(i, self.print_subplot,f)
		f.close()
		
			
			
		
class ResizingCanvas(tk.Canvas):
	def __init__(self,parent,**kwargs):
		tk.Canvas.__init__(self,parent,**kwargs)
		#print self.winfo_reqwidth(),self.winfo_reqheight() #>>>854, 404
		self.bind("<Configure>", self.on_resize)

	def on_resize(self,event):
		self.width = event.width   #>>>854
		self.height = event.height #>>>404
		self.config(width=self.width, height=self.height)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import stat_functions as stat
import loglikelihood as logl
import output
import sys
from tkinter import _tkinter




def lnsrch(ll, direction,mp,its,incr,po):
	rmsg=''
	args=ll.args_v
	g = direction.g
	dx = direction.dx
	panel=direction.panel
	LL0=ll.LL
	constr=direction.constr
	
	
	if np.sum(g*dx)<0:
		dx=-dx
		rmsg="convex function at evaluation point, direction reversed - "	
	
	ll,msg,lmbda,ok=lnsrch_master(args, dx,panel,constr,mp,rmsg,LL0)
	if  ll.LL/panel.NT<-1e+15:
		if not printout_func(0.95,'The maximization agorithm has gone mad. Resetting the argument to initial values',ll,its,direction,incr,po):return ll,msg,lmbda,ok	
		ll=logl.LL(panel.args.args_restricted,panel)
		return ll,'The maximization agorithm has gone mad. Resetting the argument to initial values',0,False
	if not ok:
		if not printout_func(0.95,msg,ll,its,direction,incr,po):return ll,msg,lmbda,ok	
		i=np.argsort(np.abs(dx))[-1]
		dx=-ll.args_v*(np.arange(len(g))==i)
		ll,msg,lmbda,ok=lnsrch_master(args, dx,panel,constr,mp,rmsg,LL0)
		if not ok:
			dx=ll.args_v*(np.arange(len(g))==i)
			ll,msg,lmbda,ok=lnsrch_master(args, dx,panel,constr,mp,rmsg,LL0)			
	return ll,msg,lmbda,ok
		
		
def solve_square_func(f0,l0,f05,l05,f1,l1,default=None):
	try:
		b=-f0*(l05+l1)/((l0-l05)*(l0-l1))
		b-=f05*(l0+l1)/((l05-l0)*(l05-l1))
		b-=f1*(l05+l0)/((l1-l05)*(l1-l0))
		c=((f0-f1)/(l0-l1)) + ((f05-f1)/(l1-l05))
		c=c/(l0-l05)
		if c<0 and b>0:#concave increasing function
			return -b/(2*c)
		else:
			return default
	except:
		return default
	
	
def lnsrch_master(args, dx,panel,constr,mp,rmsg,LL0):
	mp.send_dict_by_file({'constr':constr})
	start=0
	end=1.5
	msg=''
	for i in range(4):
		delta=(end-start)/(mp.master.cpu_count-1)
		res=get_likelihoods(args, dx, panel, constr, mp,delta,start)
		if i==0:
			res0=res[0]		
		if (res[0,1]==0 and i==0) or (res[0,0]<=res0[0] and i>0) or np.isnan(res[0,0]):#best was no change
			start=delta/mp.master.cpu_count
			end=delta
		else:
			if i>0:
				msg=f'Found increment at {round(res[0,2],4)} of Newton step'
			break
	if i>0:
		res=np.append([res0],res,0)#ensuring the original is in the LL set
		srt=np.argsort(res[:,0])[::-1]
		res=res[srt]
	res=remove_nan(res)
	if LL0>res[0,0]:
		raise RuntimeWarning('Best linsearch is poorer than the starting point. You may have discovered a bug, please notify espen.sirnes@uit.no')
	try:
		lmda=solve_square_func(res[0,0], res[0,2],res[1,0], res[1,2],res[2,0], res[2,2],res[0,2])
		ll=logl.LL(args+lmda*dx,panel,constr)
		if ll.LL<res[0,0]:
			raise RuntimeError('Something wrong with ll. You may have discovered a bug, please notify espen.sirnes@uit.no')
	except:
		ll, lmda = mp.remote_recieve(f'f{res[0,1]}',res[0,1]), res[0,2]
	if lmda==0:
		return ll,rmsg+'No increase in linesearch',0,False
	if msg=='':
		msg=f"Linesearch success ({round(lmda,6)} of Newton step)"
	return ll,rmsg+msg,lmda,True

def remove_nan(res):
	r=[]
	for i in res:
		if not np.isnan(i[0]):
			r.append(i)
	return np.array(r)	
	
				
def get_likelihoods(args, dx,panel,constr,mp,delta,start):
	expr=[]	
	lamdas=[]
	args_lst=[]
	ids=range(mp.master.cpu_count)
	for i in ids:
		lmda=start+i*delta
		a=list(args+lmda*dx)
		lamdas.append(lmda)
		args_lst.append(a)
		expr.append([f"""
try:
	f{i}=lgl.LL({a}, panel,constr)
	LL{i}=f{i}.LL
except:
	LL{i}=None
""", f'LL{i}'])
	d=mp.remote_execute(expr)
	res=[]
	for i in ids:
		if not d[f'LL{i}'] is None:
			res.append([d[f'LL{i}'],i,lamdas[i]])
	if len(res)==0:
		return np.array([[np.nan, np.nan, np.nan]])
	res=np.array(res,dtype='object')
	srt=np.argsort(res[:,0])[::-1]
	res=res[srt]
	return res
	

	
	
def maximize(panel,direction,mp,args,tab):
	"""Maxmizes logl.LL"""

	convergence_limit=panel.settings.convergence_limit.value[0]
	its, k, m, dx_norm,incr		=0,  0,     0,    None, 0
	H,  digits_precision    = None, 12
	msg,lmbda,newton_failed	='',	1, False
	direction.hessin_num, ll= None, None
	args_archive			= panel.input.args_archive
	ll=direction.init_ll(args)
	po=printout(tab,panel)
	if not printout_func(0.0,'Determining direction',ll,its,direction,incr,po):return ll,direction,po
	while 1:
		direction.get(ll,its,newton_failed,msg)
		LL0=ll.LL
			
		#Convergence test:
		constr_conv=np.max(np.abs(direction.dx_norm))<convergence_limit
		unconstr_conv=np.max(np.abs(direction.dx_unconstr)) < convergence_limit 
		conv=constr_conv or (unconstr_conv and its>3)
		args_archive.save(ll.args_d,conv,panel)
		if conv: 
			printout_func(1.0,"Convergence on zero gradient; maximum identified",ll,its,direction,incr,po)
			return ll,direction,po

		if not printout_func(0.95,"Linesearch",ll,its,direction,incr,po):return ll,direction,po
		ll,msg,lmbda,ok=lnsrch(ll,direction,mp,its,incr,po) 
		incr=ll.LL-LL0
		if not printout_func(1.0,msg,ll,its,direction,incr,po):return ll,direction,po
		its+=1
		
def printout_func(percent,msg,ll,its,direction,incr,po):
	po.printout(ll,its+1,direction,True,incr)	
	return direction.progress_bar(percent,msg)
		
	
def round_sign(x,n):
	"""rounds to n significant digits"""
	return round(x, -int(np.log10(abs(x)))+n-1)


def impose_OLS(ll,args_d,panel):
	beta,e=stat.OLS(panel,ll.X_st,ll.Y_st,return_e=True)
	args_d['omega'][0][0]=np.log(np.var(e*panel.included)*len(e[0])/np.sum(panel.included))
	args_d['beta'][:]=beta
	
class printout:
	def __init__(self,tab,panel,_print=True):
		self._print=_print
		self.tab=tab
		if tab is None:
			return
		self.panel=panel
		
	def printout(self,ll, its, direction, display_statistics,incr):
		if self.tab is None and self._print:
			print(ll.LL)
			return
		if not self._print:
			return
		#self.tab.ll=ll
		self.displaystats(display_statistics,ll)
				
		o=output.output(printout_format,ll, direction,self.panel.settings.robustcov_lags_statistics.value[1])
		o.add_heading(its,
					  top_header=" "*118+"constraints",
					  statistics=[['\nDependent: ',self.panel.input.y_name[0],None,"\n"],
								  ['Max condition index',direction.CI,3,'decimal']],
					  incr=incr)
		o.add_footer("Significance codes: .=0.1, *=0.05, **=0.01, ***=0.001,    |=collinear\n"
					+'\n' 
					+ ll.err_msg)
		tab_stops=o.get_tab_stops(self.tab.box.text_box.config()['font'][4])
		self.tab.box.text_box.config(tabs=tab_stops)		
		self.print(o)
		self.prtstr=o.printstring
		self.output_dict=o.dict
		
	def print(self,o):
		try:
			o.print(self.tab)
		except Exception as e:
			test1=e.args[0]=="main thread is not in main loop"
			#test2=e==_tkinter.TclError
			if test1:
				exit(0)
			else:
				raise e		
		
	def displaystats(self,display_statistics,ll):
		if display_statistics:
			process_charts=self.tab.charts
			process_charts.initialize(self.panel)
			process_charts.plot(ll)	



l=8
		#python variable name,	lenght,	  not string, display name,	  neg. values,	justification	next tab space
printout_format=[['names',		'namelen',	True,	'Variable names',			False,		'right', 		2],
				 ['args',		l,			False,	'Coef',						True,		'right', 		2],
				 ['dx_norm',	l,			False,	'direction',				True,		'right', 		2],
				 ['se_robust',	l,			False,	'SE(robust)',				True,		'right', 		3],
				 ['tstat',		l,			False,	't-stat.',					True,		'right', 		2],
				 ['tsign',		l,			False,	'sign.',					False,		'right', 		1],
				 ['sign_codes',	5,			True,	'',							False,		'left', 		1],
				 ['multicoll',	1,			True,	'',							False,		'left', 		2],
				 ['assco',		20,			True,	'collinear with',			False,		'center', 		2],
				 ['set_to',		6,			True,	'set to',					False,		'center', 		2],
				 ['cause',		50,			True,	'cause',					False,		'right', 		2]]			


#!/usr/bin/env python
# -*- coding: utf-8 -*-


#Todo: 



#capture singular matrix with test_small.csv
#make sure error in h function triggers an exeption


import numpy as np
import output
import panel
import warnings
import multi_core as mc
import loaddata
import model_parser
import maximize
import tempstore
import os
import direction as drctn
from gui import gui_output_tab


warnings.filterwarnings('error')
np.set_printoptions(suppress=True)
np.set_printoptions(precision=8)


def execute(model_string,dataframe, IDs_name, time_name,heteroscedasticity_factors,settings,window=None,exe_tab=None):

	"""optimizes LL using the optimization procedure in the maximize module"""
	if not exe_tab is None:
		if exe_tab.isrunning==False:return
	output_tab=gui_output_tab.output_tab(window,exe_tab)
	datainput=input_class(dataframe,model_string,IDs_name,time_name, settings,heteroscedasticity_factors)
	if settings.loadARIMA_GARCH.value:
		settings.pqdkm.value=datainput.args_archive.pqdkm
	mp,close_mp=mp_check(datainput,window)
	pqdkm=makelist(settings.pqdkm.value)
	for i in pqdkm:
		print(f'pqdkm={i}')
		results_obj=results(dataframe,datainput,settings,mp,output_tab,i)
		if len(pqdkm)>1:
			settings.loadargs.value=2
			settings.loadARIMA_GARCH.value=True
	if not mp is None and close_mp:
		mp.quit()
	return results_obj

def get_panel(model_string,dataframe, IDs_name, time_name,settings):
	datainput=input_class(dataframe,model_string,IDs_name,time_name, settings)
	return panel.panel(dataframe,datainput,settings)
	
def makelist(pqdkm):
	try:
		a=pqdkm[0][0]
		return pqdkm
	except:
		return  [pqdkm]
	

class input_class:
	def __init__(self,dataframe,model_string,IDs_name,time_name, settings,heteroscedasticity_factors=None,descr=None):
		
		t=type(settings.user_constraints.value)
		if t!=list and t!=tuple and (not settings.user_constraints.value is None):
			print("User user_constraints must be a list of tuples. user_constraints are not applied.")	
		if not hasattr(settings,'heteroscedasticity_factors'):
			settings.heteroscedasticity_factors=heteroscedasticity_factors
		self.tempfile=tempstore.tempfile_manager()
		model_parser.get_variables(self,dataframe,model_string,IDs_name,time_name,settings)
		self.descr=descr
		if descr==None:
			self.descr=model_string
		self.args_archive=tempstore.args_archive(self.descr, settings.loadargs.value)
		self.args=self.args_archive.args

	
	
class results:
	def __init__(self,dataframe,datainput,settings,mp,output_tab,pqdkm):
		print ("Creating panel")
		pnl=panel.panel(dataframe,datainput,settings,pqdkm)
		direction=drctn.direction(pnl,mp,output_tab)	
		self.mp=mp
		if not mp is None:
			mp.send_dict_by_file({'panel':pnl})
		self.ll,self.direction,self.printout_obj = maximize.maximize(pnl,direction,mp,pnl.args.args_init,output_tab)	
		self.panel=direction.panel


def mp_check(datainput,window):
	modules="""
global cf
global lgl
import calculus_functions as cf
import loglikelihood as lgl
"""	
	if window is None:
		mp=mc.multiprocess(datainput.tempfile,16,modules,['GARM','GARK','AMAq','AMAp'])
		return mp, True
	if window.mc is None:
		window.mc=mc.multiprocess(datainput.tempfile,16,modules,['GARM','GARK','AMAq','AMAp'])
	return window.mc,False
	


def indentify_dataset(glob,source):
	try:
		window=glob['window']
		datasets=window.right_tabs.data_tree.datasets
		for i in datasets:
			data_source=' '.join(datasets[i].source.split())
			editor_source=' '.join(source.split())
			if data_source==editor_source:
				return datasets[i]
	except:
		return False
			

		
def identify_global(globals,name):
	try:
		variable=globals[name]
	except:
		variable=None	
	return variable